package ru.dbotthepony.kbox2d.api

import ru.dbotthepony.kvector.util2d.AABB
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import ru.dbotthepony.kbox2d.collision.DynamicTree
import ru.dbotthepony.kbox2d.collision.BroadPhase

fun interface ProxyQueryCallback {
	/**
	 * Return false to terminate query.
	 */
	fun invoke(nodeId: Int, userData: Any?): Boolean
}

fun interface ProxyRayCastCallback {
	fun invoke(subInput: RayCastInput, nodeId: Int, userData: Any?): Double
}

interface IProxieable {
	/**
	 * For [BroadPhase]:
	 * Create a proxy with an initial AABB. Pairs are not reported until
	 * UpdatePairs is called.
	 *
	 * For [DynamicTree]:
	 * Create a proxy. Provide a tight fitting AABB and a userData pointer.
	 */
	fun createProxy(aabb: AABB, userData: Any?): Int

	/**
	 * For [BroadPhase]:
	 * Destroy a proxy. It is up to the client to remove any pairs.
	 *
	 * For [DynamicTree]:
	 * Destroy a proxy. This asserts if the id is invalid.
	 */
	fun destroyProxy(proxyID: Int)

	/**
	 * For [BroadPhase]:
	 * Call MoveProxy as many times as you like, then when you are done
	 * call UpdatePairs to finalized the proxy pairs (for your time step).
	 * @return true
	 *
	 * For [DynamicTree]:
	 * Move a proxy with a swepted AABB. If the proxy has moved outside of its fattened AABB,
	 * then the proxy is removed from the tree and re-inserted. Otherwise
	 * the function returns immediately.
	 * @return true if the proxy was re-inserted.
	 */
	fun moveProxy(proxyID: Int, aabb: AABB, displacement: Vector2d): Boolean

	/**
	 * Get user data from a proxy. Returns nullptr if the id is invalid.
 	 */
	fun getUserData(proxyID: Int): Any?

	/**
	 * Get the fat AABB for a proxy.
	 */
	fun getFatAABB(proxyID: Int): AABB

	/**
	 * Query an AABB for overlapping proxies. The callback class
	 * is called for each proxy that overlaps the supplied AABB.
	 *
	 * @return Whenever callback terminated query early
	 */
	fun query(aabb: AABB, callback: ProxyQueryCallback): Boolean

	/**
	 * Ray-cast against the proxies in the tree. This relies on the callback
	 * to perform a exact ray-cast in the case were the proxy contains a shape.
	 * The callback also performs the any collision filtering. This has performance
	 * roughly equal to k * log(n), where k is the number of collisions and n is the
	 * number of proxies in the tree.
	 * @param input the ray-cast input data. The ray extends from p1 to p1 + maxFraction * (p2 - p1).
	 * @param callback a callback class that is called for each proxy that is hit by the ray.
	 */
	fun rayCast(input: RayCastInput, callback: ProxyRayCastCallback)
}
