package ru.dbotthepony.kbox2d.collision

import ru.dbotthepony.kbox2d.api.*
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import ru.dbotthepony.kvector.vector.ndouble.times
import kotlin.math.absoluteValue

var b2_toiCalls = 0
	private set

var b2_toiRootIters = 0
	private set

var b2_toiMaxRootIters = 0
	private set

var b2_toiIters = 0
	private set

var b2_toiMaxIters = 0
	private set

var b2_toiTime = 0L
	private set

var b2_toiMaxTime = 0L
	private set

private data class MinSeparationResult(
	val indexA: Int,
	val indexB: Int,
	val separation: Double
)

private class SeparationFunction(
	cache: SimplexCache,
	val proxyA: DistanceProxy,
	val proxyB: DistanceProxy,
	val sweepA: Sweep,
	val sweepB: Sweep,
	t1: Double
) {
	enum class Type {
		POINTS,
		FACE_A,
		FACE_B,
	}

	val type: Type
	val localPoint: Vector2d
	val axis: Vector2d

	init {
		require(cache.count in 1 .. 2) { cache.count }

		val xfA = sweepA.getTransform(t1)
		val xfB = sweepB.getTransform(t1)

		if (cache.count == 1) {
			type = Type.POINTS
			val localPointA = proxyA.vertices[cache.indexA[0]]
			val localPointB = proxyB.vertices[cache.indexB[0]]
			val pointA = b2Mul(xfA, localPointA)
			val pointB = b2Mul(xfB, localPointB)
			axis = pointB - pointA

			localPoint = Vector2d.ZERO
		} else if (cache.indexA[0] == cache.indexA[1]) {
			// Two points on B and one on A.
			type = Type.FACE_B
			val localPointB1 = proxyB.vertices[cache.indexB[0]]
			val localPointB2 = proxyB.vertices[cache.indexB[1]]

			val axis = b2Cross(localPointB2 - localPointB1, 1.0).normalized
			val normal = b2Mul(xfB.q, axis)

			localPoint = 0.5 * (localPointB1 + localPointB2)
			val pointB = b2Mul(xfB, localPoint)

			val localPointA = proxyA.vertices[cache.indexA[0]]
			val pointA = b2Mul(xfA, localPointA)

			val s = b2Dot(pointA - pointB, normal)

			if (s < 0.0) {
				this.axis = -axis
			} else {
				this.axis = axis
			}
		} else {
			// Two points on A and one or two points on B.
			type = Type.FACE_A;
			val localPointA1 = proxyA.vertices[cache.indexA[0]]
			val localPointA2 = proxyA.vertices[cache.indexA[1]]

			val axis = b2Cross(localPointA2 - localPointA1, 1.0).normalized
			val normal = b2Mul(xfA.q, axis)

			localPoint = 0.5 * (localPointA1 + localPointA2)
			val pointA = b2Mul(xfA, localPoint)

			val localPointB = proxyB.vertices[cache.indexB[0]]
			val pointB = b2Mul(xfB, localPointB)

			val s = b2Dot(pointB - pointA, normal)

			if (s < 0.0) {
				this.axis = -axis
			} else {
				this.axis = axis
			}
		}
	}

	fun findMinSeparation(t: Double): MinSeparationResult {
		val xfA = sweepA.getTransform(t)
		val xfB = sweepB.getTransform(t)

		when (type) {
			Type.POINTS -> {
				val axisA = xfA.q.timesT( axis)
				val axisB = xfB.q.timesT(-axis)

				val indexA = proxyA.getSupport(axisA)
				val indexB = proxyB.getSupport(axisB)

				val localPointA = proxyA.vertices[indexA]
				val localPointB = proxyB.vertices[indexB]

				val pointA = xfA.times(localPointA)
				val pointB = xfB.times(localPointB)

				return MinSeparationResult(
					indexA = indexA,
					indexB = indexB,
					separation = (pointB - pointA).dot(axis)
				)
			}

			Type.FACE_A -> {
				val normal = xfA.q.times(axis)
				val pointA = xfA.times(localPoint)

				val axisB = xfB.q.timesT(-normal)

				val indexA = -1
				val indexB = proxyB.getSupport(axisB)

				val localPointB = proxyB.vertices[indexB]
				val pointB = xfB.times(localPointB)

				val separation = (pointB - pointA).dot(normal)
				return MinSeparationResult(
					indexA = indexA,
					indexB = indexB,
					separation = separation
				)
			}

			Type.FACE_B -> {
				val normal = xfB.q.times(axis)
				val pointB = xfB.times(localPoint)

				val axisA = xfA.q.timesT(-normal)

				val indexB = -1
				val indexA = proxyA.getSupport(axisA)

				val localPointA = proxyA.vertices[indexA]
				val pointA = xfA.times(localPointA)

				val separation = (pointA - pointB).dot(normal)
				return MinSeparationResult(
					indexA = indexA,
					indexB = indexB,
					separation = separation
				)
			}
		}
	}

	fun evaluate(indexA: Int, indexB: Int, t: Double): Double {
		val xfA = sweepA.getTransform(t)
		val xfB = sweepB.getTransform(t)

		when (type) {
			Type.POINTS -> {
				val localPointA = proxyA.vertices[indexA]
				val localPointB = proxyB.vertices[indexB]

				val pointA = xfA.times(localPointA)
				val pointB = xfB.times(localPointB)

				return (pointB - pointA).dot(axis)
			}

			Type.FACE_A -> {
				val normal = xfA.q.times(axis);
				val pointA = xfA.times(localPoint);

				val localPointB = proxyB.vertices[indexB]
				val pointB = xfB.times(localPointB)

				return (pointB - pointA).dot(normal)
			}

			Type.FACE_B -> {
				val normal = xfB.q.times(axis)
				val pointB = xfB.times(localPoint)

				val localPointA = proxyA.vertices[indexA]
				val pointA = xfA.times(localPointA)

				return (pointA - pointB).dot(normal)
			}
		}
	}
}

const val k_maxIterations = 20

/**
 * Compute the upper bound on time before two shapes penetrate. Time is represented as
 * a fraction between [0,tMax]. This uses a swept separating axis and may miss some intermediate,
 * non-tunneling collisions. If you change the time interval, you should call this function
 * again.
 *
 * Note: use [b2Distance] to compute the contact point and normal at the time of impact.
 */
fun b2TimeOfImpact(
	proxyA: DistanceProxy,
	proxyB: DistanceProxy,
	sweepA: Sweep,
	sweepB: Sweep,
	tMax: Double,		// defines sweep interval [0, tMax]
): TOIOutput {
	val timer = System.nanoTime()

	b2_toiCalls++

	var state = TOIOutput.State.UNKNOWN
	var t = tMax

	@Suppress("name_shadowing")
	val sweepA = sweepA.copy()

	@Suppress("name_shadowing")
	val sweepB = sweepB.copy()

	sweepA.normalize()
	sweepB.normalize()

	val totalRadius = proxyA.radius + proxyB.radius
	val target = b2_linearSlop.coerceAtLeast(totalRadius - 3.0 * b2_linearSlop)
	val tolerance = 0.25 * b2_linearSlop

	check(target > tolerance) { "$target <= $tolerance" }

	var t1 = 0.0
	var iter = 0

	// Prepare input for distance query.
	var cache = SimplexCache()
	val distanceInput = DistanceInput(
		proxyA = proxyA,
		proxyB = proxyB,
		useRadii = false
	)

	// The outer loop progressively attempts to compute new separating axes.
	// This loop terminates when an axis is repeated (no progress is made).
	while (true) {
		val xfA = sweepA.getTransform(t1)
		val xfB = sweepB.getTransform(t1)

		// Get the distance between shapes. We can also use the results
		// to get a separating axis.
		distanceInput.transformA = xfA
		distanceInput.transformB = xfB

		val distanceOutput = b2Distance(cache, distanceInput)
		cache = distanceOutput.newCache

		// If the shapes are overlapped, we give up on continuous collision.
		if (distanceOutput.distance <= 0.0) {
			// Failure!
			state = TOIOutput.State.OVERLAPPED
			t = 0.0
			break
		}

		// Initialize the separating axis.
		val fcn = SeparationFunction(cache, proxyA, proxyB, sweepA, sweepB, t1)

		// Compute the TOI on the separating axis. We do this by successively
		// resolving the deepest point. This loop is bounded by the number of vertices.
		var done = false
		var t2 = tMax
		var pushBackIter = 0

		while (true) {
			// Find the deepest point at t2. Store the witness point indices.
			var (indexA, indexB, s2) = fcn.findMinSeparation(t2)

			// Is the final configuration separated?
			if (s2 > target + tolerance) {
				// Victory!
				state = TOIOutput.State.SEPARATED
				t = tMax
				done = true
				break
			}

			// Has the separation reached tolerance?
			if (s2 > target - tolerance) {
				// Advance the sweeps
				t1 = t2
				break
			}

			// Compute the initial separation of the witness points.
			var s1 = fcn.evaluate(indexA, indexB, t1)

			// Check for initial overlap. This might happen if the root finder
			// runs out of iterations.
			if (s1 < target - tolerance) {
				state = TOIOutput.State.FAILED
				t = t1
				done = true
				break
			}

			// Check for touching
			if (s1 <= target + tolerance) {
				// Victory! t1 should hold the TOI (could be 0.0).
				state = TOIOutput.State.TOUCHING
				t = t1
				done = true
				break
			}

			// Compute 1D root of: f(x) - target = 0
			var rootIterCount = 0
			var a1 = t1
			var a2 = t2

			while (rootIterCount < 50) {
				// Use a mix of the secant rule and bisection.
				val t: Double

				if (rootIterCount and 1 == 1) {
					// Secant rule to improve convergence.
					t = a1 + (target - s1) * (a2 - a1) / (s2 - s1)
				} else {
					// Bisection to guarantee progress.
					t = 0.5 * (a1 + a2)
				}

				rootIterCount++
				b2_toiRootIters++

				val s = fcn.evaluate(indexA, indexB, t)

				if ((s - target).absoluteValue < tolerance) {
					// t2 holds a tentative value for t1
					t2 = t
					break
				}

				// Ensure we continue to bracket the root.
				if (s > target) {
					a1 = t
					s1 = s
				} else {
					a2 = t
					s2 = s
				}
			}

			b2_toiMaxRootIters = b2_toiMaxRootIters.coerceAtLeast(rootIterCount)
			pushBackIter++

			if (pushBackIter == b2_maxPolygonVertices) {
				break
			}
		}

		iter++
		b2_toiIters++

		if (done) {
			break
		}

		if (iter == k_maxIterations) {
			// Root finder got stuck. Semi-victory.
			state = TOIOutput.State.FAILED
			t = t1
			break
		}
	}

	b2_toiMaxIters = b2_toiMaxIters.coerceAtLeast(iter)
	val spent = System.nanoTime() - timer
	b2_toiMaxTime = b2_toiMaxTime.coerceAtLeast(spent)
	b2_toiTime += spent

	return TOIOutput(state, t)
}
