package ru.dbotthepony.kbox2d.dynamics

import ru.dbotthepony.kbox2d.api.*
import ru.dbotthepony.kbox2d.collision.BroadPhase
import ru.dbotthepony.kbox2d.collision.shapes.PolygonShape
import ru.dbotthepony.kvector.util2d.AABB
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import java.util.*
import kotlin.collections.ArrayList

/**
 * A fixture is used to attach a shape to a body for collision detection. A fixture
 * inherits its transform from its parent. Fixtures hold additional non-geometric data
 * such as friction, collision filters, etc.
 * Fixtures are created via b2Body::CreateFixture.
 * @warning you cannot reuse fixtures.
 */
class B2Fixture(
	body: B2Body,
	def: FixtureDef
) {
	/**
	 * Get the parent body of this fixture. This is nullptr if the fixture is not attached.
	 * @return the parent body.
	 */
	var body: B2Body? = body
		private set

	/**
	 * Get the next fixture in the parent body's fixture list.
	 * @return the next shape.
	 */
	var next: B2Fixture? = null
		internal set

	/**
	 * Get the previous fixture in the parent body's fixture list.
	 * @return the next shape.
	 */
	var prev: B2Fixture? = null
		internal set

	/**
	 * Get the user data that was assigned in the fixture definition. Use this to
	 * store your application specific data.
	 */
	var userData: Any? = def.userData

	/**
	 * Get the coefficient of friction.
	 * Set the coefficient of friction. This will _not_ change the friction of
	 * existing contacts.
	 */
	var friction: Double = def.friction

	/**
	 * Get the coefficient of restitution.
	 * Set the coefficient of restitution. This will _not_ change the restitution of
	 * existing contacts.
	 */
	var restitution: Double = def.restitution

	/**
	 * Get the restitution velocity threshold.
	 * Set the restitution threshold. This will _not_ change the restitution threshold of
	 * existing contacts.
	 */
	var restitutionThreshold: Double = def.restitutionThreshold

	/**
	 * Get the type of the child shape. You can use this to down cast to the concrete shape.
	 * @return the shape type.
	 */
	val type: IShape.Type get() = shape.type

	/**
	 * Set if this fixture is a sensor.
	 * Is this fixture a sensor (non-solid)?
	 * @return the true if the shape is a sensor.
	 */
	var isSensor: Boolean = def.isSensor
		set(value) {
			if (field != value) {
				checkNotNull(body) { "Already destroyed" }.isAwake = true
				field = value
			}
		}

	/**
	 * Get the child shape. You can modify the child shape, however you should not change the
	 * number of vertices because this will crash some collision caching mechanisms.
	 * Manipulating the shape may lead to non-physical behavior.
	 */
	val shape: IShape<*> = requireNotNull(def.shape) { "null shape provided" }.copy()

	init {
		if (shape is PolygonShape) {
			shape.validate(true)
		}

		if (shape.childCount < 0) {
			throw IllegalArgumentException("Shape $shape has ${shape.childCount} children")
		}
	}

	private val internalProxies = ArrayList<FixtureProxy>(shape.childCount)
	val proxies: List<FixtureProxy> = Collections.unmodifiableList(internalProxies)

	/**
	 * Set the density of this fixture. This will _not_ automatically adjust the mass
	 * of the body. You must call b2Body::ResetMassData to update the body's mass.
	 * Get the density of this fixture.
	 */
	var density: Double = def.density
		set(value) {
			require(value.isFinite()) { "Infinite density" }
			require(!value.isNaN()) { "NaN density" }
			require(value >= 0.0) { "Negative density of $value" }
			field = value
		}

	/**
	 * Set the contact filtering data. This will not update contacts until the next time
	 * step when either parent body is active and awake.
	 * This automatically calls Refilter.
	 */
	var filter: IFilter = def.filter.immutable()
		set(value) {
			if (value is ImmutableFilter)
				field = value
			else
				field = (value as Filter).immutable()

			refilter()
		}

	/**
	 * **KBox2D extension**.
	 *
	 * Convenience function for quickly destroying this fixture.
	 */
	fun destroy() {
		checkNotNull(body) { "Already destroyed" }.destroyFixture(this)
	}

	internal fun unlink() {
		check(body != null) { "Already destroyed" }
		check(internalProxies.isEmpty()) { "Still having proxies" }
		body = null
	}

	/**
	 * These support body activation/deactivation.
	 */
	internal fun createProxies(broadPhase: BroadPhase, xf: Transform) {
		check(body != null) { "Already destroyed" }
		check(internalProxies.isEmpty()) { "Already having proxies" }

		for (i in 0 until shape.childCount) {
			val aabb = shape.computeAABB(xf, i)

			val proxy = FixtureProxy(
				fixture = this,
				childIndex = i,
				aabb = aabb,
			)

			proxy.proxyId = broadPhase.createProxy(aabb, proxy)
			internalProxies.add(proxy)
		}
	}

	/**
	 * These support body activation/deactivation.
	 */
	internal fun destroyProxies(broadPhase: BroadPhase) {
		check(body != null) { "Already destroyed" }

		// Destroy proxies in the broad-phase.
		for (proxy in internalProxies) {
			broadPhase.destroyProxy(proxy.proxyId)
		}

		internalProxies.clear()
	}

	internal fun synchronize(broadPhase: BroadPhase, transform1: Transform, transform2: Transform) {
		check(body != null) { "Already destroyed" }

		for (proxy in internalProxies) {
			val aabb1 = shape.computeAABB(transform1, proxy.childIndex)
			val aabb2 = shape.computeAABB(transform2, proxy.childIndex)

			proxy.aabb = aabb1.combine(aabb2)
			val displacement = aabb2.centre - aabb1.centre
			broadPhase.moveProxy(proxy.proxyId, proxy.aabb, displacement)
		}
	}

	/**
	 * Call this if you want to establish collision that was previously disabled by b2ContactFilter::ShouldCollide.
	 */
	fun refilter() {
		val body = body
		check(body != null) { "Already destroyed" }
		var edge = body.contactEdge

		while (edge != null) {
			val contact = edge.contact
			val fixtureA = contact.fixtureA
			val fixtureB = contact.fixtureB

			if (fixtureA == this || fixtureB == this)
				contact.flagForFiltering()

			edge = edge.next
		}

		val world = body.world
		val broadPhase = world.contactManager.broadPhase

		for (proxy in internalProxies) {
			broadPhase.touchProxy(proxy.proxyId)
		}
	}

	/**
	 * Test a point for containment in this fixture.
	 * @param p a point in world coordinates.
	 */
	fun testPoint(p: Vector2d): Boolean {
		return shape.testPoint(checkNotNull(body) { "Tried to use detached fixture" }.transform, p)
	}

	/**
	 * Cast a ray against this shape.
	 * @param output the ray-cast results.
	 * @param input the ray-cast input parameters.
	 * @param childIndex the child shape index (e.g. edge index)
	 */
	fun rayCast(input: RayCastInput, childIndex: Int): RayCastOutput {
		return shape.rayCast(input, checkNotNull(body) { "Tried to use detached fixture" }.transform, childIndex)
	}

	/**
	 * Get the mass data for this fixture. The mass data is based on the density and
	 * the shape. The rotational inertia is about the shape's origin. This operation
	 * may be expensive.
	 */
	fun getMassData(): MassData {
		return shape.computeMass(density)
	}

	/**
	 * Get the fixture's AABB. This AABB may be enlarge and/or stale.
	 * If you need a more accurate AABB, compute it using the shape and
	 * the body transform.
	 */
	fun getAABB(childIndex: Int): AABB {
		return proxies[childIndex].aabb
	}

	/**
	 * Dump this fixture to the log file.
	 */
	fun dump(childIndex: Int) { }
}
