package ru.dbotthepony.kbox2d.dynamics.joint

import ru.dbotthepony.kbox2d.api.*
import ru.dbotthepony.kbox2d.dynamics.B2Body
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import ru.dbotthepony.kvector.vector.ndouble.times

// Gear Joint:
// C0 = (coordinate1 + ratio * coordinate2)_initial
// C = (coordinate1 + ratio * coordinate2) - C0 = 0
// J = [J1 ratio * J2]
// K = J * invM * JT
//   = J1 * invM1 * J1T + ratio * ratio * J2 * invM2 * J2T
//
// Revolute:
// coordinate = rotation
// Cdot = angularVelocity
// J = [0 0 1]
// K = J * invM * JT = invI
//
// Prismatic:
// coordinate = dot(p - pg, ug)
// Cdot = dot(v + cross(w, r), ug)
// J = [ug cross(r, ug)]
// K = J * invM * JT = invMass + invI * cross(r, ug)^2

/**
 * A gear joint is used to connect two joints together. Either joint
 * can be a revolute or prismatic joint. You specify a gear ratio
 * to bind the motions together:
 * coordinate1 + ratio * coordinate2 = constant
 * The ratio can be negative or positive. If one joint is a revolute joint
 * and the other joint is a prismatic joint, then the ratio will have units
 * of length or units of 1/length.
 * @warning You have to manually destroy the gear joint if joint1 or joint2
 * is destroyed.
 */
class GearJoint(def: GearJointDef) : AbstractJoint(def) {
	// KBox2D: Due to multiple local variables names clashing with class fields
	// this class contain hard reference to this in methods to clear confusion
	init {
		require(def.joint1 != def.joint2) { "Definition specify the same joint ${def.joint1}" }
	}

	/**
	 * Get the first joint.
	 */
	val joint1: AbstractJoint = def.joint1

	/**
	 * Get the second joint.
	 */
	val joint2: AbstractJoint = def.joint2

	val typeA: JointType = def.joint1.type
	val typeB: JointType = def.joint2.type

	init {
		check(typeA == JointType.REVOLUTE || typeA == JointType.PRISMATIC) { "Invalid joint A: $typeA" }
		check(typeB == JointType.REVOLUTE || typeB == JointType.PRISMATIC) { "Invalid joint B: $typeB" }
	}

	// Solver shared
	private var localAnchorA: Vector2d
	private var localAnchorB: Vector2d
	private var localAnchorC: Vector2d
	private var localAnchorD: Vector2d

	private var localAxisC: Vector2d
	private var localAxisD: Vector2d

	private var referenceAngleA: Double
	private var referenceAngleB: Double

	private var constant: Double

	/**
	 * Set/Get the gear ratio.
	 */
	var ratio: Double = def.ratio
		set(value) {
			if (!value.isFinite()) {
				throw IllegalArgumentException("Tried to set infinite ratio")
			}

			if (value.isNaN()) {
				throw IllegalArgumentException("Tried to set NaN ratio")
			}

			field = value
		}

	private var tolerance: Double

	private var impulse: Double = 0.0

	// Solver temp
	private var indexA: Int = 0
	private var indexB: Int = 0
	private var indexC: Int = 0
	private var indexD: Int = 0

	private var lcA: Vector2d = Vector2d.ZERO
	private var lcB: Vector2d = Vector2d.ZERO
	private var lcC: Vector2d = Vector2d.ZERO
	private var lcD: Vector2d = Vector2d.ZERO

	private var mA: Double = 0.0
	private var mB: Double = 0.0
	private var mC: Double = 0.0
	private var mD: Double = 0.0
	private var iA: Double = 0.0
	private var iB: Double = 0.0
	private var iC: Double = 0.0
	private var iD: Double = 0.0

	private var JvAC: Vector2d = Vector2d.ZERO
	private var JvBD: Vector2d = Vector2d.ZERO

	private var JwA: Double = 0.0
	private var JwB: Double = 0.0
	private var JwC: Double = 0.0
	private var JwD: Double = 0.0

	private var mass: Double = 0.0

	// Body A is connected to body C
	// Body B is connected to body D
	private var _bodyC: B2Body? = joint1.bodyA
	private var _bodyD: B2Body?

	override fun unlink() {
		super.unlink()
		_bodyC = null
		_bodyD = null
	}

	val bodyC: B2Body get() = checkNotNull(_bodyC) { "Body C is not present" }
	val bodyD: B2Body get() = checkNotNull(_bodyD) { "Body D is not present" }

	init {
		_bodyA = joint1.bodyB

		val coordinateA: Double
		val coordinateB: Double

		// Body B on joint1 must be dynamic
		check(bodyA.type == BodyType.DYNAMIC) { "Body A is expected to be DYNAMIC, hot ${bodyA.type}" }

		// Get geometry of joint1
		val xfA = bodyA.transform
		val aA = bodyA.sweep.a
		val xfC = bodyC.transform
		val aC = bodyC.sweep.a

		if (typeA == JointType.REVOLUTE) {
			val revolute = def.joint1 as RevoluteJoint

			localAnchorC = revolute.localAnchorA
			localAnchorA = revolute.localAnchorB
			referenceAngleA = revolute.referenceAngle
			localAxisC = Vector2d.ZERO

			coordinateA = aA - aC - referenceAngleA

			// position error is measured in radians
			tolerance = b2_angularSlop
		} else {
			val prismatic = def.joint1 as PrismaticJoint
			localAnchorC = prismatic.localAnchorA
			localAnchorA = prismatic.localAnchorB
			referenceAngleA = prismatic.referenceAngle
			localAxisC = prismatic.localXAxisA

			val pC = localAnchorC
			val pA = b2MulT(xfC.q, b2Mul(xfA.q, localAnchorA) + (xfA.p - xfC.p))
			coordinateA = b2Dot(pA - pC, localAxisC)

			// position error is measured in meters
			tolerance = b2_linearSlop
		}

		_bodyD = joint2.bodyA
		_bodyB = joint2.bodyB

		// Body B on joint2 must be dynamic
		check(bodyB.type == BodyType.DYNAMIC) { "Body is expected to be DYNAMIC, got ${bodyB.type}"}

		// Get geometry of joint2
		val xfB = bodyB.transform
		val aB = bodyB.sweep.a
		val xfD = bodyD.transform
		val aD = bodyD.sweep.a

		if (typeB == JointType.REVOLUTE) {
			val revolute = def.joint2 as RevoluteJoint
			localAnchorD = revolute.localAnchorA
			localAnchorB = revolute.localAnchorB
			referenceAngleB = revolute.referenceAngle
			localAxisD = Vector2d.ZERO

			coordinateB = aB - aD - referenceAngleB
		} else {
			val prismatic = def.joint2 as PrismaticJoint
			localAnchorD = prismatic.localAnchorA
			localAnchorB = prismatic.localAnchorB
			referenceAngleB = prismatic.referenceAngle
			localAxisD = prismatic.localXAxisA

			val pD = localAnchorD
			val pB = b2MulT(xfD.q, b2Mul(xfB.q, localAnchorB) + (xfB.p - xfD.p))
			coordinateB = b2Dot(pB - pD, localAxisD)
		}

		constant = coordinateA + ratio * coordinateB
	}

	override fun initVelocityConstraints(data: B2SolverData) {
		if (!joint1.isValid || !joint2.isValid) {
			throw IllegalStateException("$this is orphaned!")
		}

		this.indexA = this.bodyA.islandIndex
		this.indexB = this.bodyB.islandIndex
		this.indexC = this.bodyC.islandIndex
		this.indexD = this.bodyD.islandIndex
		this.lcA = this.bodyA.sweep.localCenter
		this.lcB = this.bodyB.sweep.localCenter
		this.lcC = this.bodyC.sweep.localCenter
		this.lcD = this.bodyD.sweep.localCenter
		this.mA = this.bodyA.invMass
		this.mB = this.bodyB.invMass
		this.mC = this.bodyC.invMass
		this.mD = this.bodyD.invMass
		this.iA = this.bodyA.rotInertiaInv
		this.iB = this.bodyB.rotInertiaInv
		this.iC = this.bodyC.rotInertiaInv
		this.iD = this.bodyD.rotInertiaInv

		val aA = data.positions[this.indexA].a
		var vA = data.velocities[this.indexA].v
		var wA = data.velocities[this.indexA].w

		val aB = data.positions[this.indexB].a
		var vB = data.velocities[this.indexB].v
		var wB = data.velocities[this.indexB].w

		val aC = data.positions[this.indexC].a
		var vC = data.velocities[this.indexC].v
		var wC = data.velocities[this.indexC].w

		val aD = data.positions[this.indexD].a
		var vD = data.velocities[this.indexD].v
		var wD = data.velocities[this.indexD].w

		val qA = Rotation(aA)
		val qB = Rotation(aB)
		val qC = Rotation(aC)
		val qD = Rotation(aD)

		this.mass = 0.0

		if (this.typeA == JointType.REVOLUTE) {
			this.JvAC = Vector2d.ZERO
			this.JwA = 1.0
			this.JwC = 1.0
			this.mass += this.iA + this.iC
		} else {
			val u = b2Mul(qC, this.localAxisC)
			val rC = b2Mul(qC, this.localAnchorC - this.lcC)
			val rA = b2Mul(qA, this.localAnchorA - this.lcA)
			this.JvAC = u
			this.JwC = b2Cross(rC, u)
			this.JwA = b2Cross(rA, u)
			this.mass += this.mC + this.mA + this.iC * this.JwC * this.JwC + this.iA * this.JwA * this.JwA
		}

		if (this.typeB == JointType.REVOLUTE) {
			this.JvBD = Vector2d.ZERO
			this.JwB = this.ratio
			this.JwD = this.ratio
			this.mass += this.ratio * this.ratio * (this.iB + this.iD)
		} else {
			val u = b2Mul(qD, this.localAxisD)
			val rD = b2Mul(qD, this.localAnchorD - this.lcD)
			val rB = b2Mul(qB, this.localAnchorB - this.lcB)
			this.JvBD = this.ratio * u
			this.JwD = this.ratio * b2Cross(rD, u)
			this.JwB = this.ratio * b2Cross(rB, u)
			this.mass += this.ratio * this.ratio * (this.mD + this.mB) + this.iD * this.JwD * this.JwD + this.iB * this.JwB * this.JwB
		}

		// Compute effective mass.
		this.mass = if (this.mass > 0.0) 1.0 / this.mass else 0.0

		if (data.step.warmStarting) {
			vA += (this.mA * this.impulse) * this.JvAC
			wA += this.iA * this.impulse * this.JwA
			vB += (this.mB * this.impulse) * this.JvBD
			wB += this.iB * this.impulse * this.JwB
			vC -= (this.mC * this.impulse) * this.JvAC
			wC -= this.iC * this.impulse * this.JwC
			vD -= (this.mD * this.impulse) * this.JvBD
			wD -= this.iD * this.impulse * this.JwD
		} else {
			this.impulse = 0.0
		}

		data.velocities[this.indexA].v = vA
		data.velocities[this.indexA].w = wA
		data.velocities[this.indexB].v = vB
		data.velocities[this.indexB].w = wB
		data.velocities[this.indexC].v = vC
		data.velocities[this.indexC].w = wC
		data.velocities[this.indexD].v = vD
		data.velocities[this.indexD].w = wD
	}

	override fun solveVelocityConstraints(data: B2SolverData) {
		var vA = data.velocities[this.indexA].v
		var wA = data.velocities[this.indexA].w
		var vB = data.velocities[this.indexB].v
		var wB = data.velocities[this.indexB].w
		var vC = data.velocities[this.indexC].v
		var wC = data.velocities[this.indexC].w
		var vD = data.velocities[this.indexD].v
		var wD = data.velocities[this.indexD].w

		var Cdot = b2Dot(this.JvAC, vA - vC) + b2Dot(this.JvBD, vB - vD)
		Cdot += (this.JwA * wA - this.JwC * wC) + (this.JwB * wB - this.JwD * wD)

		val impulse = -this.mass * Cdot
		this.impulse += impulse

		vA += (this.mA * impulse) * this.JvAC
		wA += this.iA * impulse * this.JwA
		vB += (this.mB * impulse) * this.JvBD
		wB += this.iB * impulse * this.JwB
		vC -= (this.mC * impulse) * this.JvAC
		wC -= this.iC * impulse * this.JwC
		vD -= (this.mD * impulse) * this.JvBD
		wD -= this.iD * impulse * this.JwD

		data.velocities[this.indexA].v = vA
		data.velocities[this.indexA].w = wA
		data.velocities[this.indexB].v = vB
		data.velocities[this.indexB].w = wB
		data.velocities[this.indexC].v = vC
		data.velocities[this.indexC].w = wC
		data.velocities[this.indexD].v = vD
		data.velocities[this.indexD].w = wD
	}

	override fun solvePositionConstraints(data: B2SolverData): Boolean {
		var cA = data.positions[this.indexA].c
		var aA = data.positions[this.indexA].a
		var cB = data.positions[this.indexB].c
		var aB = data.positions[this.indexB].a
		var cC = data.positions[this.indexC].c
		var aC = data.positions[this.indexC].a
		var cD = data.positions[this.indexD].c
		var aD = data.positions[this.indexD].a

		val qA = Rotation(aA)
		val qB = Rotation(aB)
		val qC = Rotation(aC)
		val qD = Rotation(aD)

		val coordinateA: Double
		val coordinateB: Double

		val JvAC: Vector2d
		val JvBD: Vector2d
		val JwA: Double
		val JwB: Double
		val JwC: Double
		val JwD: Double

		var mass = 0.0

		if (this.typeA == JointType.REVOLUTE) {
			JvAC = Vector2d.ZERO
			JwA = 1.0
			JwC = 1.0
			mass += this.iA + this.iC

			coordinateA = aA - aC - this.referenceAngleA
		} else {
			val u = b2Mul(qC, this.localAxisC)
			val rC = b2Mul(qC, this.localAnchorC - this.lcC)
			val rA = b2Mul(qA, this.localAnchorA - this.lcA)
			JvAC = u
			JwC = b2Cross(rC, u)
			JwA = b2Cross(rA, u)
			mass += this.mC + this.mA + this.iC * JwC * JwC + this.iA * JwA * JwA

			val pC = this.localAnchorC - this.lcC
			val pA = b2MulT(qC, rA + (cA - cC))
			coordinateA = b2Dot(pA - pC, this.localAxisC)
		}

		if (this.typeB == JointType.REVOLUTE) {
			JvBD = Vector2d.ZERO
			JwB = this.ratio
			JwD = this.ratio
			mass += this.ratio * this.ratio * (this.iB + this.iD)

			coordinateB = aB - aD - this.referenceAngleB
		} else {
			val u = b2Mul(qD, this.localAxisD)
			val rD = b2Mul(qD, this.localAnchorD - this.lcD)
			val rB = b2Mul(qB, this.localAnchorB - this.lcB)
			JvBD = this.ratio * u
			JwD = this.ratio * b2Cross(rD, u)
			JwB = this.ratio * b2Cross(rB, u)
			mass += this.ratio * this.ratio * (this.mD + this.mB) + this.iD * JwD * JwD + this.iB * JwB * JwB

			val pD = this.localAnchorD - this.lcD
			val pB = b2MulT(qD, rB + (cB - cD))
			coordinateB = b2Dot(pB - pD, this.localAxisD)
		}

		val C = (coordinateA + this.ratio * coordinateB) - this.constant

		var impulse = 0.0

		if (mass > 0.0f) {
			impulse = -C / mass
		}

		cA += this.mA * impulse * JvAC
		aA += this.iA * impulse * JwA
		cB += this.mB * impulse * JvBD
		aB += this.iB * impulse * JwB
		cC -= this.mC * impulse * JvAC
		aC -= this.iC * impulse * JwC
		cD -= this.mD * impulse * JvBD
		aD -= this.iD * impulse * JwD

		data.positions[this.indexA].c = cA
		data.positions[this.indexA].a = aA
		data.positions[this.indexB].c = cB
		data.positions[this.indexB].a = aB
		data.positions[this.indexC].c = cC
		data.positions[this.indexC].a = aC
		data.positions[this.indexD].c = cD
		data.positions[this.indexD].a = aD

		if (b2Abs(C) < this.tolerance) {
			return true
		}

		return false
	}

	override val anchorA: Vector2d
		get() = bodyA.getWorldPoint(localAnchorA)
	override val anchorB: Vector2d
		get() = bodyB.getWorldPoint(localAnchorB)

	override fun getReactionForce(inv_dt: Double): Vector2d {
		return inv_dt * (impulse * JvAC)
	}

	override fun getReactionTorque(inv_dt: Double): Double {
		return inv_dt * (impulse * JwA)
	}
}
