
@file:Suppress("unused")

package ru.dbotthepony.kvector.matrix.nfloat

import ru.dbotthepony.kvector.api.concrete.IMatrix2f
import ru.dbotthepony.kvector.narray.Float2Dimensional
import ru.dbotthepony.kvector.vector.nfloat.Vector2f
import ru.dbotthepony.kvector.matrix.ndouble.Matrix4d
import ru.dbotthepony.kvector.matrix.ndouble.MutableMatrix4d
import ru.dbotthepony.kvector.vector.nfloat.MutableVector2f

/**
 * Represents immutable concrete matrix with 2x2 dimensions, storing values as [Float]s.
 *
 * See [Matrix4d] for documentation, with reflection of this class.
 *
 * Vectorized access use [Vector2f], generic interface is represented by [IMatrix2f].
 */
class Matrix2f : AbstractMatrixVf<Matrix2f>, IMatrix2f<Matrix2f> {
	constructor(
		c00: Float = 1f, c01: Float = 0f,
		c10: Float = 0f, c11: Float = 1f,
	) : super(2, 2, Float2Dimensional(2, 2).also {
		it[0, 0] = c00
		it[1, 0] = c10

		it[0, 1] = c01
		it[1, 1] = c11
	})

	private constructor(memory: Float2Dimensional) : super(2, 2, memory)

	override val flexible: Boolean = false

	override fun factorize(memory: Float2Dimensional): Matrix2f {
		requireSizeEquals(memory)
		return Matrix2f(memory)
	}

	override val c00: Float get() = memory[0, 0]
	override val c10: Float get() = memory[1, 0]
	override val c01: Float get() = memory[0, 1]
	override val c11: Float get() = memory[1, 1]

	override val r00 by this::c00
	override val r01 by this::c10
	override val r10 by this::c01
	override val r11 by this::c11

	override val a11 by this::c00
	override val a21 by this::c01
	override val a12 by this::c10
	override val a22 by this::c11

	override val b11 by this::c00
	override val b21 by this::c10
	override val b12 by this::c01
	override val b22 by this::c11

	override val c0: Vector2f get() = Vector2f(c00, c01)
	override val c1: Vector2f get() = Vector2f(c10, c11)

	override val r0: Vector2f get() = Vector2f(r00, r01)
	override val r1: Vector2f get() = Vector2f(r10, r11)

	companion object {
		/**
		 * Initializes new matrix, with arguments taken in *Row-Major* order, that is, fist row of values
		 * is fist row in created matrix.
		 */
		fun rm(
			r00: Float = 1f, r01: Float = 0f,
			r10: Float = 0f, r11: Float = 1f,
		): Matrix2f {
			return Matrix2f(
				c00 = r00, c10 = r01,
				c01 = r10, c11 = r11,
			)
		}

		/**
		 * Initializes new zero matrix.
		 *
		 * Since [Matrix2f] is immutable, it is encouraged you use [ZERO] property instead.
		 */
		fun zero(): Matrix2f {
			return Matrix2f(
				c00 = 0f, c10 = 0f,
				c01 = 0f, c11 = 0f,
			)
		}

		/**
		 * All values of this matrix are zero
		 */
		@JvmField val ZERO = zero()

		/**
		 * Main diagonal of this matrix is 1s
		 */
		@JvmField val IDENTITY = Matrix2f()
	}
}

/**
 * Represents mutable concrete matrix with 2x2 dimensions, storing values as [Double]s.
 *
 * See [MutableMatrix4d] for documentation, with reflection of this class.
 *
 * Vectorized access use [MutableVector2f], generic interface is represented by [IMatrix2f].
 */
class MutableMatrix2f : AbstractMutableMatrixVf<MutableMatrix2f>, IMatrix2f<MutableMatrix2f> {
	constructor(
		c00: Float = 1f, c01: Float = 0f,
		c10: Float = 0f, c11: Float = 1f,
	) : super(2, 2, Float2Dimensional(2, 2).also {
		it[0, 0] = c00
		it[1, 0] = c10

		it[0, 1] = c01
		it[1, 1] = c11
	})

	private constructor(memory: Float2Dimensional) : super(2, 2, memory)

	override val flexible: Boolean = false

	override fun factorize(memory: Float2Dimensional): MutableMatrix2f {
		requireSizeEquals(memory)
		return MutableMatrix2f(memory)
	}


	override var c00: Float
		get() = memory[0, 0]
		set(value) { memory[0, 0] = value }
	override var c10: Float
		get() = memory[1, 0]
		set(value) { memory[1, 0] = value }
	override var c01: Float
		get() = memory[0, 1]
		set(value) { memory[0, 1] = value }
	override var c11: Float
		get() = memory[1, 1]
		set(value) { memory[1, 1] = value }

	override var r00 by this::c00
	override var r01 by this::c10
	override var r10 by this::c01
	override var r11 by this::c11

	override var a11 by this::c00
	override var a21 by this::c01
	override var a12 by this::c10
	override var a22 by this::c11

	override var b11 by this::c00
	override var b21 by this::c10
	override var b12 by this::c01
	override var b22 by this::c11

	override var c0: Vector2f
		get() = object : MutableVector2f(c00, c01) {
			override var x by this@MutableMatrix2f::c00
			override var y by this@MutableMatrix2f::c01
		}
		set(value) {
			c00 = value.x
			c01 = value.y
		}
	override var c1: Vector2f
		get() = object : MutableVector2f(c10, c11) {
			override var x by this@MutableMatrix2f::c10
			override var y by this@MutableMatrix2f::c11
		}
		set(value) {
			c10 = value.x
			c11 = value.y
		}

	override var r0: Vector2f
		get() = object : MutableVector2f(r00, r01) {
			override var x by this@MutableMatrix2f::r00
			override var y by this@MutableMatrix2f::r01
		}
		set(value) {
			r00 = value.x
			r01 = value.y
		}
	override var r1: Vector2f
		get() = object : MutableVector2f(r10, r11) {
			override var x by this@MutableMatrix2f::r10
			override var y by this@MutableMatrix2f::r11
		}
		set(value) {
			r10 = value.x
			r11 = value.y
		}

	companion object {
		/**
		 * Initializes new matrix, with arguments taken in *Row-Major* order, that is, fist row of values
		 * is fist row in created matrix.
		 */
		fun rm(
			r00: Float = 1f, r01: Float = 0f,
			r10: Float = 0f, r11: Float = 1f,
		): MutableMatrix2f {
			return MutableMatrix2f(
				c00 = r00, c10 = r01,
				c01 = r10, c11 = r11,
			)
		}

		/**
		 * Initializes new zero matrix.
		 */
		fun zero(): MutableMatrix2f {
			return MutableMatrix2f(
				c00 = 0f, c10 = 0f,
				c01 = 0f, c11 = 0f,
			)
		}
	}
}
