
@file:Suppress("nothing_to_inline", "unused")

package ru.dbotthepony.kvector.vector.nint

import ru.dbotthepony.kvector.api.*
import kotlin.math.absoluteValue

/**
 * 4D Vector, representing four-dimensional coordinates as [Int]s
 */
open class Vector4i(
	open val x: Int = 0,
	open val y: Int = 0,
	open val z: Int = 0,
	open val w: Int = 0,
) : AbstractVector<Vector4i>(), IWholeVector<Vector4i>, IIntVector<Vector4i>, IStruct4i, IMatrixGetterInt {
	constructor(input: IStruct2i, z: Int = 0, w: Int = 0) : this(input.component1(), input.component2(), z, w)
	constructor(x: Int, input: IStruct2i, w: Int = 0) : this(x, input.component1(), input.component2(), w)
	constructor(x: Int, y: Int, input: IStruct2i) : this(x, y, input.component1(), input.component2())

	constructor(input: IStruct3i, w: Int = 0) : this(input.component1(), input.component2(), input.component3(), w)
	constructor(x: Int, input: IStruct3i) : this(x, input.component1(), input.component2(), input.component3())
	constructor(input: IStruct4i) : this(input.component1(), input.component2(), input.component3(), input.component4())

	override fun get(column: Int, row: Int): Int {
		if (column != 0) {
			throw IndexOutOfBoundsException("Vectors are 1 column matrices ($column given)")
		}

		return when (row) {
			0 -> x
			1 -> y
			2 -> z
			3 -> w
			else -> throw IndexOutOfBoundsException("Row out of bounds: $row")
		}
	}

	final override val rows: Int = 3

	final override fun component1() = x
	final override fun component2() = y
	final override fun component3() = z
	final override fun component4() = w

	open val r by this::x
	open val g by this::y
	open val b by this::z
	open val a by this::w

	open val s by this::x
	open val t by this::y
	open val p by this::z
	open val q by this::w

	final override val lengthSquared: Double
		get() = (x.toLong() * x.toLong() + y.toLong() * y.toLong() + z.toLong() * z.toLong()).toDouble()
	final override val isFinite: Boolean = true
	final override val isNaN: Boolean = false

	override fun plus(other: Vector4i): Vector4i {
		return Vector4i(
			x + other.x,
			y + other.y,
			z + other.z,
			w + other.w,
		)
	}

	override fun minus(other: Vector4i): Vector4i {
		return Vector4i(
			x - other.x,
			y - other.y,
			z - other.z,
			w - other.w,
		)
	}

	override fun times(other: Vector4i): Vector4i {
		return Vector4i(
			x * other.x,
			y * other.y,
			z * other.z,
			w * other.w,
		)
	}

	override fun div(other: Vector4i): Vector4i {
		return Vector4i(
			x / other.x,
			y / other.y,
			z / other.z,
			w / other.w,
		)
	}

	override val absoluteValue: Vector4i
		get() = Vector4i(x.absoluteValue, y.absoluteValue, z.absoluteValue, w.absoluteValue)

	override fun coerceAtMost(maximal: Vector4i): Vector4i {
		return Vector4i(
			x.coerceAtMost(maximal.x),
			y.coerceAtMost(maximal.y),
			z.coerceAtMost(maximal.z),
			w.coerceAtMost(maximal.w),
		)
	}

	override fun coerceAtLeast(minimal: Vector4i): Vector4i {
		return Vector4i(
			x.coerceAtLeast(minimal.x),
			y.coerceAtLeast(minimal.y),
			z.coerceAtLeast(minimal.z),
			w.coerceAtLeast(minimal.w),
		)
	}

	override fun clamp(minimal: Vector4i, maximal: Vector4i): Vector4i {
		return Vector4i(
			x.coerceAtLeast(minimal.x).coerceAtMost(maximal.x),
			y.coerceAtLeast(minimal.y).coerceAtMost(maximal.y),
			z.coerceAtLeast(minimal.z).coerceAtMost(maximal.z),
			w.coerceAtLeast(minimal.w).coerceAtMost(maximal.w),
		)
	}

	override fun unaryMinus(): Vector4i {
		return Vector4i(-x, -y, -z, -w)
	}

	override fun distanceSquared(other: Vector4i): Double {
		val x = x.toDouble() - other.x.toDouble()
		val y = y.toDouble() - other.y.toDouble()
		val z = z.toDouble() - other.z.toDouble()
		val w = w.toDouble() - other.w.toDouble()

		return x * x + y * y + z * z + w * w
	}

	override fun wholeDistanceSquared(other: Vector4i): Long {
		val x = x.toLong() - other.x.toLong()
		val y = y.toLong() - other.y.toLong()
		val z = z.toLong() - other.z.toLong()
		val w = w.toLong() - other.w.toLong()

		return x * x + y * y + z * z + w * w
	}

	override fun times(other: Int): Vector4i {
		return Vector4i(
			x * other,
			y * other,
			z * other,
			w * other,
		)
	}

	override fun div(other: Int): Vector4i {
		return Vector4i(
			x / other,
			y / other,
			z / other,
			w / other,
		)
	}

	override fun equals(other: Any?): Boolean {
		if (this === other) return true
		if (javaClass != other?.javaClass) return false

		other as Vector4i

		if (x != other.x) return false
		if (y != other.y) return false
		if (z != other.z) return false
		if (w != other.w) return false

		return true
	}

	override fun hashCode(): Int {
		var result = x
		result = 31 * result + y
		result = 31 * result + z
		result = 31 * result + w
		return result
	}

	override fun toString(): String {
		return "[${x}i ${y}i ${z}i ${w}i]"
	}

	inline val xx get() = Vector2i(x, x)
	inline val xy get() = Vector2i(x, y)
	inline val xz get() = Vector2i(x, z)
	inline val xw get() = Vector2i(x, w)
	inline val yx get() = Vector2i(y, x)
	inline val yy get() = Vector2i(y, y)
	inline val yz get() = Vector2i(y, z)
	inline val yw get() = Vector2i(y, w)
	inline val zx get() = Vector2i(z, x)
	inline val zy get() = Vector2i(z, y)
	inline val zz get() = Vector2i(z, z)
	inline val zw get() = Vector2i(z, w)
	inline val wx get() = Vector2i(w, x)
	inline val wy get() = Vector2i(w, y)
	inline val wz get() = Vector2i(w, z)
	inline val ww get() = Vector2i(w, w)
	inline val xxx get() = Vector3i(x, x, x)
	inline val xxy get() = Vector3i(x, x, y)
	inline val xxz get() = Vector3i(x, x, z)
	inline val xxw get() = Vector3i(x, x, w)
	inline val xyx get() = Vector3i(x, y, x)
	inline val xyy get() = Vector3i(x, y, y)
	inline val xyz get() = Vector3i(x, y, z)
	inline val xyw get() = Vector3i(x, y, w)
	inline val xzx get() = Vector3i(x, z, x)
	inline val xzy get() = Vector3i(x, z, y)
	inline val xzz get() = Vector3i(x, z, z)
	inline val xzw get() = Vector3i(x, z, w)
	inline val xwx get() = Vector3i(x, w, x)
	inline val xwy get() = Vector3i(x, w, y)
	inline val xwz get() = Vector3i(x, w, z)
	inline val xww get() = Vector3i(x, w, w)
	inline val yxx get() = Vector3i(y, x, x)
	inline val yxy get() = Vector3i(y, x, y)
	inline val yxz get() = Vector3i(y, x, z)
	inline val yxw get() = Vector3i(y, x, w)
	inline val yyx get() = Vector3i(y, y, x)
	inline val yyy get() = Vector3i(y, y, y)
	inline val yyz get() = Vector3i(y, y, z)
	inline val yyw get() = Vector3i(y, y, w)
	inline val yzx get() = Vector3i(y, z, x)
	inline val yzy get() = Vector3i(y, z, y)
	inline val yzz get() = Vector3i(y, z, z)
	inline val yzw get() = Vector3i(y, z, w)
	inline val ywx get() = Vector3i(y, w, x)
	inline val ywy get() = Vector3i(y, w, y)
	inline val ywz get() = Vector3i(y, w, z)
	inline val yww get() = Vector3i(y, w, w)
	inline val zxx get() = Vector3i(z, x, x)
	inline val zxy get() = Vector3i(z, x, y)
	inline val zxz get() = Vector3i(z, x, z)
	inline val zxw get() = Vector3i(z, x, w)
	inline val zyx get() = Vector3i(z, y, x)
	inline val zyy get() = Vector3i(z, y, y)
	inline val zyz get() = Vector3i(z, y, z)
	inline val zyw get() = Vector3i(z, y, w)
	inline val zzx get() = Vector3i(z, z, x)
	inline val zzy get() = Vector3i(z, z, y)
	inline val zzz get() = Vector3i(z, z, z)
	inline val zzw get() = Vector3i(z, z, w)
	inline val zwx get() = Vector3i(z, w, x)
	inline val zwy get() = Vector3i(z, w, y)
	inline val zwz get() = Vector3i(z, w, z)
	inline val zww get() = Vector3i(z, w, w)
	inline val wxx get() = Vector3i(w, x, x)
	inline val wxy get() = Vector3i(w, x, y)
	inline val wxz get() = Vector3i(w, x, z)
	inline val wxw get() = Vector3i(w, x, w)
	inline val wyx get() = Vector3i(w, y, x)
	inline val wyy get() = Vector3i(w, y, y)
	inline val wyz get() = Vector3i(w, y, z)
	inline val wyw get() = Vector3i(w, y, w)
	inline val wzx get() = Vector3i(w, z, x)
	inline val wzy get() = Vector3i(w, z, y)
	inline val wzz get() = Vector3i(w, z, z)
	inline val wzw get() = Vector3i(w, z, w)
	inline val wwx get() = Vector3i(w, w, x)
	inline val wwy get() = Vector3i(w, w, y)
	inline val wwz get() = Vector3i(w, w, z)
	inline val www get() = Vector3i(w, w, w)
	inline val xxxx get() = Vector4i(x, x, x, x)
	inline val xxxy get() = Vector4i(x, x, x, y)
	inline val xxxz get() = Vector4i(x, x, x, z)
	inline val xxxw get() = Vector4i(x, x, x, w)
	inline val xxyx get() = Vector4i(x, x, y, x)
	inline val xxyy get() = Vector4i(x, x, y, y)
	inline val xxyz get() = Vector4i(x, x, y, z)
	inline val xxyw get() = Vector4i(x, x, y, w)
	inline val xxzx get() = Vector4i(x, x, z, x)
	inline val xxzy get() = Vector4i(x, x, z, y)
	inline val xxzz get() = Vector4i(x, x, z, z)
	inline val xxzw get() = Vector4i(x, x, z, w)
	inline val xxwx get() = Vector4i(x, x, w, x)
	inline val xxwy get() = Vector4i(x, x, w, y)
	inline val xxwz get() = Vector4i(x, x, w, z)
	inline val xxww get() = Vector4i(x, x, w, w)
	inline val xyxx get() = Vector4i(x, y, x, x)
	inline val xyxy get() = Vector4i(x, y, x, y)
	inline val xyxz get() = Vector4i(x, y, x, z)
	inline val xyxw get() = Vector4i(x, y, x, w)
	inline val xyyx get() = Vector4i(x, y, y, x)
	inline val xyyy get() = Vector4i(x, y, y, y)
	inline val xyyz get() = Vector4i(x, y, y, z)
	inline val xyyw get() = Vector4i(x, y, y, w)
	inline val xyzx get() = Vector4i(x, y, z, x)
	inline val xyzy get() = Vector4i(x, y, z, y)
	inline val xyzz get() = Vector4i(x, y, z, z)
	inline val xyzw get() = Vector4i(x, y, z, w)
	inline val xywx get() = Vector4i(x, y, w, x)
	inline val xywy get() = Vector4i(x, y, w, y)
	inline val xywz get() = Vector4i(x, y, w, z)
	inline val xyww get() = Vector4i(x, y, w, w)
	inline val xzxx get() = Vector4i(x, z, x, x)
	inline val xzxy get() = Vector4i(x, z, x, y)
	inline val xzxz get() = Vector4i(x, z, x, z)
	inline val xzxw get() = Vector4i(x, z, x, w)
	inline val xzyx get() = Vector4i(x, z, y, x)
	inline val xzyy get() = Vector4i(x, z, y, y)
	inline val xzyz get() = Vector4i(x, z, y, z)
	inline val xzyw get() = Vector4i(x, z, y, w)
	inline val xzzx get() = Vector4i(x, z, z, x)
	inline val xzzy get() = Vector4i(x, z, z, y)
	inline val xzzz get() = Vector4i(x, z, z, z)
	inline val xzzw get() = Vector4i(x, z, z, w)
	inline val xzwx get() = Vector4i(x, z, w, x)
	inline val xzwy get() = Vector4i(x, z, w, y)
	inline val xzwz get() = Vector4i(x, z, w, z)
	inline val xzww get() = Vector4i(x, z, w, w)
	inline val xwxx get() = Vector4i(x, w, x, x)
	inline val xwxy get() = Vector4i(x, w, x, y)
	inline val xwxz get() = Vector4i(x, w, x, z)
	inline val xwxw get() = Vector4i(x, w, x, w)
	inline val xwyx get() = Vector4i(x, w, y, x)
	inline val xwyy get() = Vector4i(x, w, y, y)
	inline val xwyz get() = Vector4i(x, w, y, z)
	inline val xwyw get() = Vector4i(x, w, y, w)
	inline val xwzx get() = Vector4i(x, w, z, x)
	inline val xwzy get() = Vector4i(x, w, z, y)
	inline val xwzz get() = Vector4i(x, w, z, z)
	inline val xwzw get() = Vector4i(x, w, z, w)
	inline val xwwx get() = Vector4i(x, w, w, x)
	inline val xwwy get() = Vector4i(x, w, w, y)
	inline val xwwz get() = Vector4i(x, w, w, z)
	inline val xwww get() = Vector4i(x, w, w, w)
	inline val yxxx get() = Vector4i(y, x, x, x)
	inline val yxxy get() = Vector4i(y, x, x, y)
	inline val yxxz get() = Vector4i(y, x, x, z)
	inline val yxxw get() = Vector4i(y, x, x, w)
	inline val yxyx get() = Vector4i(y, x, y, x)
	inline val yxyy get() = Vector4i(y, x, y, y)
	inline val yxyz get() = Vector4i(y, x, y, z)
	inline val yxyw get() = Vector4i(y, x, y, w)
	inline val yxzx get() = Vector4i(y, x, z, x)
	inline val yxzy get() = Vector4i(y, x, z, y)
	inline val yxzz get() = Vector4i(y, x, z, z)
	inline val yxzw get() = Vector4i(y, x, z, w)
	inline val yxwx get() = Vector4i(y, x, w, x)
	inline val yxwy get() = Vector4i(y, x, w, y)
	inline val yxwz get() = Vector4i(y, x, w, z)
	inline val yxww get() = Vector4i(y, x, w, w)
	inline val yyxx get() = Vector4i(y, y, x, x)
	inline val yyxy get() = Vector4i(y, y, x, y)
	inline val yyxz get() = Vector4i(y, y, x, z)
	inline val yyxw get() = Vector4i(y, y, x, w)
	inline val yyyx get() = Vector4i(y, y, y, x)
	inline val yyyy get() = Vector4i(y, y, y, y)
	inline val yyyz get() = Vector4i(y, y, y, z)
	inline val yyyw get() = Vector4i(y, y, y, w)
	inline val yyzx get() = Vector4i(y, y, z, x)
	inline val yyzy get() = Vector4i(y, y, z, y)
	inline val yyzz get() = Vector4i(y, y, z, z)
	inline val yyzw get() = Vector4i(y, y, z, w)
	inline val yywx get() = Vector4i(y, y, w, x)
	inline val yywy get() = Vector4i(y, y, w, y)
	inline val yywz get() = Vector4i(y, y, w, z)
	inline val yyww get() = Vector4i(y, y, w, w)
	inline val yzxx get() = Vector4i(y, z, x, x)
	inline val yzxy get() = Vector4i(y, z, x, y)
	inline val yzxz get() = Vector4i(y, z, x, z)
	inline val yzxw get() = Vector4i(y, z, x, w)
	inline val yzyx get() = Vector4i(y, z, y, x)
	inline val yzyy get() = Vector4i(y, z, y, y)
	inline val yzyz get() = Vector4i(y, z, y, z)
	inline val yzyw get() = Vector4i(y, z, y, w)
	inline val yzzx get() = Vector4i(y, z, z, x)
	inline val yzzy get() = Vector4i(y, z, z, y)
	inline val yzzz get() = Vector4i(y, z, z, z)
	inline val yzzw get() = Vector4i(y, z, z, w)
	inline val yzwx get() = Vector4i(y, z, w, x)
	inline val yzwy get() = Vector4i(y, z, w, y)
	inline val yzwz get() = Vector4i(y, z, w, z)
	inline val yzww get() = Vector4i(y, z, w, w)
	inline val ywxx get() = Vector4i(y, w, x, x)
	inline val ywxy get() = Vector4i(y, w, x, y)
	inline val ywxz get() = Vector4i(y, w, x, z)
	inline val ywxw get() = Vector4i(y, w, x, w)
	inline val ywyx get() = Vector4i(y, w, y, x)
	inline val ywyy get() = Vector4i(y, w, y, y)
	inline val ywyz get() = Vector4i(y, w, y, z)
	inline val ywyw get() = Vector4i(y, w, y, w)
	inline val ywzx get() = Vector4i(y, w, z, x)
	inline val ywzy get() = Vector4i(y, w, z, y)
	inline val ywzz get() = Vector4i(y, w, z, z)
	inline val ywzw get() = Vector4i(y, w, z, w)
	inline val ywwx get() = Vector4i(y, w, w, x)
	inline val ywwy get() = Vector4i(y, w, w, y)
	inline val ywwz get() = Vector4i(y, w, w, z)
	inline val ywww get() = Vector4i(y, w, w, w)
	inline val zxxx get() = Vector4i(z, x, x, x)
	inline val zxxy get() = Vector4i(z, x, x, y)
	inline val zxxz get() = Vector4i(z, x, x, z)
	inline val zxxw get() = Vector4i(z, x, x, w)
	inline val zxyx get() = Vector4i(z, x, y, x)
	inline val zxyy get() = Vector4i(z, x, y, y)
	inline val zxyz get() = Vector4i(z, x, y, z)
	inline val zxyw get() = Vector4i(z, x, y, w)
	inline val zxzx get() = Vector4i(z, x, z, x)
	inline val zxzy get() = Vector4i(z, x, z, y)
	inline val zxzz get() = Vector4i(z, x, z, z)
	inline val zxzw get() = Vector4i(z, x, z, w)
	inline val zxwx get() = Vector4i(z, x, w, x)
	inline val zxwy get() = Vector4i(z, x, w, y)
	inline val zxwz get() = Vector4i(z, x, w, z)
	inline val zxww get() = Vector4i(z, x, w, w)
	inline val zyxx get() = Vector4i(z, y, x, x)
	inline val zyxy get() = Vector4i(z, y, x, y)
	inline val zyxz get() = Vector4i(z, y, x, z)
	inline val zyxw get() = Vector4i(z, y, x, w)
	inline val zyyx get() = Vector4i(z, y, y, x)
	inline val zyyy get() = Vector4i(z, y, y, y)
	inline val zyyz get() = Vector4i(z, y, y, z)
	inline val zyyw get() = Vector4i(z, y, y, w)
	inline val zyzx get() = Vector4i(z, y, z, x)
	inline val zyzy get() = Vector4i(z, y, z, y)
	inline val zyzz get() = Vector4i(z, y, z, z)
	inline val zyzw get() = Vector4i(z, y, z, w)
	inline val zywx get() = Vector4i(z, y, w, x)
	inline val zywy get() = Vector4i(z, y, w, y)
	inline val zywz get() = Vector4i(z, y, w, z)
	inline val zyww get() = Vector4i(z, y, w, w)
	inline val zzxx get() = Vector4i(z, z, x, x)
	inline val zzxy get() = Vector4i(z, z, x, y)
	inline val zzxz get() = Vector4i(z, z, x, z)
	inline val zzxw get() = Vector4i(z, z, x, w)
	inline val zzyx get() = Vector4i(z, z, y, x)
	inline val zzyy get() = Vector4i(z, z, y, y)
	inline val zzyz get() = Vector4i(z, z, y, z)
	inline val zzyw get() = Vector4i(z, z, y, w)
	inline val zzzx get() = Vector4i(z, z, z, x)
	inline val zzzy get() = Vector4i(z, z, z, y)
	inline val zzzz get() = Vector4i(z, z, z, z)
	inline val zzzw get() = Vector4i(z, z, z, w)
	inline val zzwx get() = Vector4i(z, z, w, x)
	inline val zzwy get() = Vector4i(z, z, w, y)
	inline val zzwz get() = Vector4i(z, z, w, z)
	inline val zzww get() = Vector4i(z, z, w, w)
	inline val zwxx get() = Vector4i(z, w, x, x)
	inline val zwxy get() = Vector4i(z, w, x, y)
	inline val zwxz get() = Vector4i(z, w, x, z)
	inline val zwxw get() = Vector4i(z, w, x, w)
	inline val zwyx get() = Vector4i(z, w, y, x)
	inline val zwyy get() = Vector4i(z, w, y, y)
	inline val zwyz get() = Vector4i(z, w, y, z)
	inline val zwyw get() = Vector4i(z, w, y, w)
	inline val zwzx get() = Vector4i(z, w, z, x)
	inline val zwzy get() = Vector4i(z, w, z, y)
	inline val zwzz get() = Vector4i(z, w, z, z)
	inline val zwzw get() = Vector4i(z, w, z, w)
	inline val zwwx get() = Vector4i(z, w, w, x)
	inline val zwwy get() = Vector4i(z, w, w, y)
	inline val zwwz get() = Vector4i(z, w, w, z)
	inline val zwww get() = Vector4i(z, w, w, w)
	inline val wxxx get() = Vector4i(w, x, x, x)
	inline val wxxy get() = Vector4i(w, x, x, y)
	inline val wxxz get() = Vector4i(w, x, x, z)
	inline val wxxw get() = Vector4i(w, x, x, w)
	inline val wxyx get() = Vector4i(w, x, y, x)
	inline val wxyy get() = Vector4i(w, x, y, y)
	inline val wxyz get() = Vector4i(w, x, y, z)
	inline val wxyw get() = Vector4i(w, x, y, w)
	inline val wxzx get() = Vector4i(w, x, z, x)
	inline val wxzy get() = Vector4i(w, x, z, y)
	inline val wxzz get() = Vector4i(w, x, z, z)
	inline val wxzw get() = Vector4i(w, x, z, w)
	inline val wxwx get() = Vector4i(w, x, w, x)
	inline val wxwy get() = Vector4i(w, x, w, y)
	inline val wxwz get() = Vector4i(w, x, w, z)
	inline val wxww get() = Vector4i(w, x, w, w)
	inline val wyxx get() = Vector4i(w, y, x, x)
	inline val wyxy get() = Vector4i(w, y, x, y)
	inline val wyxz get() = Vector4i(w, y, x, z)
	inline val wyxw get() = Vector4i(w, y, x, w)
	inline val wyyx get() = Vector4i(w, y, y, x)
	inline val wyyy get() = Vector4i(w, y, y, y)
	inline val wyyz get() = Vector4i(w, y, y, z)
	inline val wyyw get() = Vector4i(w, y, y, w)
	inline val wyzx get() = Vector4i(w, y, z, x)
	inline val wyzy get() = Vector4i(w, y, z, y)
	inline val wyzz get() = Vector4i(w, y, z, z)
	inline val wyzw get() = Vector4i(w, y, z, w)
	inline val wywx get() = Vector4i(w, y, w, x)
	inline val wywy get() = Vector4i(w, y, w, y)
	inline val wywz get() = Vector4i(w, y, w, z)
	inline val wyww get() = Vector4i(w, y, w, w)
	inline val wzxx get() = Vector4i(w, z, x, x)
	inline val wzxy get() = Vector4i(w, z, x, y)
	inline val wzxz get() = Vector4i(w, z, x, z)
	inline val wzxw get() = Vector4i(w, z, x, w)
	inline val wzyx get() = Vector4i(w, z, y, x)
	inline val wzyy get() = Vector4i(w, z, y, y)
	inline val wzyz get() = Vector4i(w, z, y, z)
	inline val wzyw get() = Vector4i(w, z, y, w)
	inline val wzzx get() = Vector4i(w, z, z, x)
	inline val wzzy get() = Vector4i(w, z, z, y)
	inline val wzzz get() = Vector4i(w, z, z, z)
	inline val wzzw get() = Vector4i(w, z, z, w)
	inline val wzwx get() = Vector4i(w, z, w, x)
	inline val wzwy get() = Vector4i(w, z, w, y)
	inline val wzwz get() = Vector4i(w, z, w, z)
	inline val wzww get() = Vector4i(w, z, w, w)
	inline val wwxx get() = Vector4i(w, w, x, x)
	inline val wwxy get() = Vector4i(w, w, x, y)
	inline val wwxz get() = Vector4i(w, w, x, z)
	inline val wwxw get() = Vector4i(w, w, x, w)
	inline val wwyx get() = Vector4i(w, w, y, x)
	inline val wwyy get() = Vector4i(w, w, y, y)
	inline val wwyz get() = Vector4i(w, w, y, z)
	inline val wwyw get() = Vector4i(w, w, y, w)
	inline val wwzx get() = Vector4i(w, w, z, x)
	inline val wwzy get() = Vector4i(w, w, z, y)
	inline val wwzz get() = Vector4i(w, w, z, z)
	inline val wwzw get() = Vector4i(w, w, z, w)
	inline val wwwx get() = Vector4i(w, w, w, x)
	inline val wwwy get() = Vector4i(w, w, w, y)
	inline val wwwz get() = Vector4i(w, w, w, z)
	inline val wwww get() = Vector4i(w, w, w, w)
	inline val rr get() = Vector2i(r, r)
	inline val rg get() = Vector2i(r, g)
	inline val rb get() = Vector2i(r, b)
	inline val ra get() = Vector2i(r, a)
	inline val gr get() = Vector2i(g, r)
	inline val gg get() = Vector2i(g, g)
	inline val gb get() = Vector2i(g, b)
	inline val ga get() = Vector2i(g, a)
	inline val br get() = Vector2i(b, r)
	inline val bg get() = Vector2i(b, g)
	inline val bb get() = Vector2i(b, b)
	inline val ba get() = Vector2i(b, a)
	inline val ar get() = Vector2i(a, r)
	inline val ag get() = Vector2i(a, g)
	inline val ab get() = Vector2i(a, b)
	inline val aa get() = Vector2i(a, a)
	inline val rrr get() = Vector3i(r, r, r)
	inline val rrg get() = Vector3i(r, r, g)
	inline val rrb get() = Vector3i(r, r, b)
	inline val rra get() = Vector3i(r, r, a)
	inline val rgr get() = Vector3i(r, g, r)
	inline val rgg get() = Vector3i(r, g, g)
	inline val rgb get() = Vector3i(r, g, b)
	inline val rga get() = Vector3i(r, g, a)
	inline val rbr get() = Vector3i(r, b, r)
	inline val rbg get() = Vector3i(r, b, g)
	inline val rbb get() = Vector3i(r, b, b)
	inline val rba get() = Vector3i(r, b, a)
	inline val rar get() = Vector3i(r, a, r)
	inline val rag get() = Vector3i(r, a, g)
	inline val rab get() = Vector3i(r, a, b)
	inline val raa get() = Vector3i(r, a, a)
	inline val grr get() = Vector3i(g, r, r)
	inline val grg get() = Vector3i(g, r, g)
	inline val grb get() = Vector3i(g, r, b)
	inline val gra get() = Vector3i(g, r, a)
	inline val ggr get() = Vector3i(g, g, r)
	inline val ggg get() = Vector3i(g, g, g)
	inline val ggb get() = Vector3i(g, g, b)
	inline val gga get() = Vector3i(g, g, a)
	inline val gbr get() = Vector3i(g, b, r)
	inline val gbg get() = Vector3i(g, b, g)
	inline val gbb get() = Vector3i(g, b, b)
	inline val gba get() = Vector3i(g, b, a)
	inline val gar get() = Vector3i(g, a, r)
	inline val gag get() = Vector3i(g, a, g)
	inline val gab get() = Vector3i(g, a, b)
	inline val gaa get() = Vector3i(g, a, a)
	inline val brr get() = Vector3i(b, r, r)
	inline val brg get() = Vector3i(b, r, g)
	inline val brb get() = Vector3i(b, r, b)
	inline val bra get() = Vector3i(b, r, a)
	inline val bgr get() = Vector3i(b, g, r)
	inline val bgg get() = Vector3i(b, g, g)
	inline val bgb get() = Vector3i(b, g, b)
	inline val bga get() = Vector3i(b, g, a)
	inline val bbr get() = Vector3i(b, b, r)
	inline val bbg get() = Vector3i(b, b, g)
	inline val bbb get() = Vector3i(b, b, b)
	inline val bba get() = Vector3i(b, b, a)
	inline val bar get() = Vector3i(b, a, r)
	inline val bag get() = Vector3i(b, a, g)
	inline val bab get() = Vector3i(b, a, b)
	inline val baa get() = Vector3i(b, a, a)
	inline val arr get() = Vector3i(a, r, r)
	inline val arg get() = Vector3i(a, r, g)
	inline val arb get() = Vector3i(a, r, b)
	inline val ara get() = Vector3i(a, r, a)
	inline val agr get() = Vector3i(a, g, r)
	inline val agg get() = Vector3i(a, g, g)
	inline val agb get() = Vector3i(a, g, b)
	inline val aga get() = Vector3i(a, g, a)
	inline val abr get() = Vector3i(a, b, r)
	inline val abg get() = Vector3i(a, b, g)
	inline val abb get() = Vector3i(a, b, b)
	inline val aba get() = Vector3i(a, b, a)
	inline val aar get() = Vector3i(a, a, r)
	inline val aag get() = Vector3i(a, a, g)
	inline val aab get() = Vector3i(a, a, b)
	inline val aaa get() = Vector3i(a, a, a)
	inline val rrrr get() = Vector4i(r, r, r, r)
	inline val rrrg get() = Vector4i(r, r, r, g)
	inline val rrrb get() = Vector4i(r, r, r, b)
	inline val rrra get() = Vector4i(r, r, r, a)
	inline val rrgr get() = Vector4i(r, r, g, r)
	inline val rrgg get() = Vector4i(r, r, g, g)
	inline val rrgb get() = Vector4i(r, r, g, b)
	inline val rrga get() = Vector4i(r, r, g, a)
	inline val rrbr get() = Vector4i(r, r, b, r)
	inline val rrbg get() = Vector4i(r, r, b, g)
	inline val rrbb get() = Vector4i(r, r, b, b)
	inline val rrba get() = Vector4i(r, r, b, a)
	inline val rrar get() = Vector4i(r, r, a, r)
	inline val rrag get() = Vector4i(r, r, a, g)
	inline val rrab get() = Vector4i(r, r, a, b)
	inline val rraa get() = Vector4i(r, r, a, a)
	inline val rgrr get() = Vector4i(r, g, r, r)
	inline val rgrg get() = Vector4i(r, g, r, g)
	inline val rgrb get() = Vector4i(r, g, r, b)
	inline val rgra get() = Vector4i(r, g, r, a)
	inline val rggr get() = Vector4i(r, g, g, r)
	inline val rggg get() = Vector4i(r, g, g, g)
	inline val rggb get() = Vector4i(r, g, g, b)
	inline val rgga get() = Vector4i(r, g, g, a)
	inline val rgbr get() = Vector4i(r, g, b, r)
	inline val rgbg get() = Vector4i(r, g, b, g)
	inline val rgbb get() = Vector4i(r, g, b, b)
	inline val rgba get() = Vector4i(r, g, b, a)
	inline val rgar get() = Vector4i(r, g, a, r)
	inline val rgag get() = Vector4i(r, g, a, g)
	inline val rgab get() = Vector4i(r, g, a, b)
	inline val rgaa get() = Vector4i(r, g, a, a)
	inline val rbrr get() = Vector4i(r, b, r, r)
	inline val rbrg get() = Vector4i(r, b, r, g)
	inline val rbrb get() = Vector4i(r, b, r, b)
	inline val rbra get() = Vector4i(r, b, r, a)
	inline val rbgr get() = Vector4i(r, b, g, r)
	inline val rbgg get() = Vector4i(r, b, g, g)
	inline val rbgb get() = Vector4i(r, b, g, b)
	inline val rbga get() = Vector4i(r, b, g, a)
	inline val rbbr get() = Vector4i(r, b, b, r)
	inline val rbbg get() = Vector4i(r, b, b, g)
	inline val rbbb get() = Vector4i(r, b, b, b)
	inline val rbba get() = Vector4i(r, b, b, a)
	inline val rbar get() = Vector4i(r, b, a, r)
	inline val rbag get() = Vector4i(r, b, a, g)
	inline val rbab get() = Vector4i(r, b, a, b)
	inline val rbaa get() = Vector4i(r, b, a, a)
	inline val rarr get() = Vector4i(r, a, r, r)
	inline val rarg get() = Vector4i(r, a, r, g)
	inline val rarb get() = Vector4i(r, a, r, b)
	inline val rara get() = Vector4i(r, a, r, a)
	inline val ragr get() = Vector4i(r, a, g, r)
	inline val ragg get() = Vector4i(r, a, g, g)
	inline val ragb get() = Vector4i(r, a, g, b)
	inline val raga get() = Vector4i(r, a, g, a)
	inline val rabr get() = Vector4i(r, a, b, r)
	inline val rabg get() = Vector4i(r, a, b, g)
	inline val rabb get() = Vector4i(r, a, b, b)
	inline val raba get() = Vector4i(r, a, b, a)
	inline val raar get() = Vector4i(r, a, a, r)
	inline val raag get() = Vector4i(r, a, a, g)
	inline val raab get() = Vector4i(r, a, a, b)
	inline val raaa get() = Vector4i(r, a, a, a)
	inline val grrr get() = Vector4i(g, r, r, r)
	inline val grrg get() = Vector4i(g, r, r, g)
	inline val grrb get() = Vector4i(g, r, r, b)
	inline val grra get() = Vector4i(g, r, r, a)
	inline val grgr get() = Vector4i(g, r, g, r)
	inline val grgg get() = Vector4i(g, r, g, g)
	inline val grgb get() = Vector4i(g, r, g, b)
	inline val grga get() = Vector4i(g, r, g, a)
	inline val grbr get() = Vector4i(g, r, b, r)
	inline val grbg get() = Vector4i(g, r, b, g)
	inline val grbb get() = Vector4i(g, r, b, b)
	inline val grba get() = Vector4i(g, r, b, a)
	inline val grar get() = Vector4i(g, r, a, r)
	inline val grag get() = Vector4i(g, r, a, g)
	inline val grab get() = Vector4i(g, r, a, b)
	inline val graa get() = Vector4i(g, r, a, a)
	inline val ggrr get() = Vector4i(g, g, r, r)
	inline val ggrg get() = Vector4i(g, g, r, g)
	inline val ggrb get() = Vector4i(g, g, r, b)
	inline val ggra get() = Vector4i(g, g, r, a)
	inline val gggr get() = Vector4i(g, g, g, r)
	inline val gggg get() = Vector4i(g, g, g, g)
	inline val gggb get() = Vector4i(g, g, g, b)
	inline val ggga get() = Vector4i(g, g, g, a)
	inline val ggbr get() = Vector4i(g, g, b, r)
	inline val ggbg get() = Vector4i(g, g, b, g)
	inline val ggbb get() = Vector4i(g, g, b, b)
	inline val ggba get() = Vector4i(g, g, b, a)
	inline val ggar get() = Vector4i(g, g, a, r)
	inline val ggag get() = Vector4i(g, g, a, g)
	inline val ggab get() = Vector4i(g, g, a, b)
	inline val ggaa get() = Vector4i(g, g, a, a)
	inline val gbrr get() = Vector4i(g, b, r, r)
	inline val gbrg get() = Vector4i(g, b, r, g)
	inline val gbrb get() = Vector4i(g, b, r, b)
	inline val gbra get() = Vector4i(g, b, r, a)
	inline val gbgr get() = Vector4i(g, b, g, r)
	inline val gbgg get() = Vector4i(g, b, g, g)
	inline val gbgb get() = Vector4i(g, b, g, b)
	inline val gbga get() = Vector4i(g, b, g, a)
	inline val gbbr get() = Vector4i(g, b, b, r)
	inline val gbbg get() = Vector4i(g, b, b, g)
	inline val gbbb get() = Vector4i(g, b, b, b)
	inline val gbba get() = Vector4i(g, b, b, a)
	inline val gbar get() = Vector4i(g, b, a, r)
	inline val gbag get() = Vector4i(g, b, a, g)
	inline val gbab get() = Vector4i(g, b, a, b)
	inline val gbaa get() = Vector4i(g, b, a, a)
	inline val garr get() = Vector4i(g, a, r, r)
	inline val garg get() = Vector4i(g, a, r, g)
	inline val garb get() = Vector4i(g, a, r, b)
	inline val gara get() = Vector4i(g, a, r, a)
	inline val gagr get() = Vector4i(g, a, g, r)
	inline val gagg get() = Vector4i(g, a, g, g)
	inline val gagb get() = Vector4i(g, a, g, b)
	inline val gaga get() = Vector4i(g, a, g, a)
	inline val gabr get() = Vector4i(g, a, b, r)
	inline val gabg get() = Vector4i(g, a, b, g)
	inline val gabb get() = Vector4i(g, a, b, b)
	inline val gaba get() = Vector4i(g, a, b, a)
	inline val gaar get() = Vector4i(g, a, a, r)
	inline val gaag get() = Vector4i(g, a, a, g)
	inline val gaab get() = Vector4i(g, a, a, b)
	inline val gaaa get() = Vector4i(g, a, a, a)
	inline val brrr get() = Vector4i(b, r, r, r)
	inline val brrg get() = Vector4i(b, r, r, g)
	inline val brrb get() = Vector4i(b, r, r, b)
	inline val brra get() = Vector4i(b, r, r, a)
	inline val brgr get() = Vector4i(b, r, g, r)
	inline val brgg get() = Vector4i(b, r, g, g)
	inline val brgb get() = Vector4i(b, r, g, b)
	inline val brga get() = Vector4i(b, r, g, a)
	inline val brbr get() = Vector4i(b, r, b, r)
	inline val brbg get() = Vector4i(b, r, b, g)
	inline val brbb get() = Vector4i(b, r, b, b)
	inline val brba get() = Vector4i(b, r, b, a)
	inline val brar get() = Vector4i(b, r, a, r)
	inline val brag get() = Vector4i(b, r, a, g)
	inline val brab get() = Vector4i(b, r, a, b)
	inline val braa get() = Vector4i(b, r, a, a)
	inline val bgrr get() = Vector4i(b, g, r, r)
	inline val bgrg get() = Vector4i(b, g, r, g)
	inline val bgrb get() = Vector4i(b, g, r, b)
	inline val bgra get() = Vector4i(b, g, r, a)
	inline val bggr get() = Vector4i(b, g, g, r)
	inline val bggg get() = Vector4i(b, g, g, g)
	inline val bggb get() = Vector4i(b, g, g, b)
	inline val bgga get() = Vector4i(b, g, g, a)
	inline val bgbr get() = Vector4i(b, g, b, r)
	inline val bgbg get() = Vector4i(b, g, b, g)
	inline val bgbb get() = Vector4i(b, g, b, b)
	inline val bgba get() = Vector4i(b, g, b, a)
	inline val bgar get() = Vector4i(b, g, a, r)
	inline val bgag get() = Vector4i(b, g, a, g)
	inline val bgab get() = Vector4i(b, g, a, b)
	inline val bgaa get() = Vector4i(b, g, a, a)
	inline val bbrr get() = Vector4i(b, b, r, r)
	inline val bbrg get() = Vector4i(b, b, r, g)
	inline val bbrb get() = Vector4i(b, b, r, b)
	inline val bbra get() = Vector4i(b, b, r, a)
	inline val bbgr get() = Vector4i(b, b, g, r)
	inline val bbgg get() = Vector4i(b, b, g, g)
	inline val bbgb get() = Vector4i(b, b, g, b)
	inline val bbga get() = Vector4i(b, b, g, a)
	inline val bbbr get() = Vector4i(b, b, b, r)
	inline val bbbg get() = Vector4i(b, b, b, g)
	inline val bbbb get() = Vector4i(b, b, b, b)
	inline val bbba get() = Vector4i(b, b, b, a)
	inline val bbar get() = Vector4i(b, b, a, r)
	inline val bbag get() = Vector4i(b, b, a, g)
	inline val bbab get() = Vector4i(b, b, a, b)
	inline val bbaa get() = Vector4i(b, b, a, a)
	inline val barr get() = Vector4i(b, a, r, r)
	inline val barg get() = Vector4i(b, a, r, g)
	inline val barb get() = Vector4i(b, a, r, b)
	inline val bara get() = Vector4i(b, a, r, a)
	inline val bagr get() = Vector4i(b, a, g, r)
	inline val bagg get() = Vector4i(b, a, g, g)
	inline val bagb get() = Vector4i(b, a, g, b)
	inline val baga get() = Vector4i(b, a, g, a)
	inline val babr get() = Vector4i(b, a, b, r)
	inline val babg get() = Vector4i(b, a, b, g)
	inline val babb get() = Vector4i(b, a, b, b)
	inline val baba get() = Vector4i(b, a, b, a)
	inline val baar get() = Vector4i(b, a, a, r)
	inline val baag get() = Vector4i(b, a, a, g)
	inline val baab get() = Vector4i(b, a, a, b)
	inline val baaa get() = Vector4i(b, a, a, a)
	inline val arrr get() = Vector4i(a, r, r, r)
	inline val arrg get() = Vector4i(a, r, r, g)
	inline val arrb get() = Vector4i(a, r, r, b)
	inline val arra get() = Vector4i(a, r, r, a)
	inline val argr get() = Vector4i(a, r, g, r)
	inline val argg get() = Vector4i(a, r, g, g)
	inline val argb get() = Vector4i(a, r, g, b)
	inline val arga get() = Vector4i(a, r, g, a)
	inline val arbr get() = Vector4i(a, r, b, r)
	inline val arbg get() = Vector4i(a, r, b, g)
	inline val arbb get() = Vector4i(a, r, b, b)
	inline val arba get() = Vector4i(a, r, b, a)
	inline val arar get() = Vector4i(a, r, a, r)
	inline val arag get() = Vector4i(a, r, a, g)
	inline val arab get() = Vector4i(a, r, a, b)
	inline val araa get() = Vector4i(a, r, a, a)
	inline val agrr get() = Vector4i(a, g, r, r)
	inline val agrg get() = Vector4i(a, g, r, g)
	inline val agrb get() = Vector4i(a, g, r, b)
	inline val agra get() = Vector4i(a, g, r, a)
	inline val aggr get() = Vector4i(a, g, g, r)
	inline val aggg get() = Vector4i(a, g, g, g)
	inline val aggb get() = Vector4i(a, g, g, b)
	inline val agga get() = Vector4i(a, g, g, a)
	inline val agbr get() = Vector4i(a, g, b, r)
	inline val agbg get() = Vector4i(a, g, b, g)
	inline val agbb get() = Vector4i(a, g, b, b)
	inline val agba get() = Vector4i(a, g, b, a)
	inline val agar get() = Vector4i(a, g, a, r)
	inline val agag get() = Vector4i(a, g, a, g)
	inline val agab get() = Vector4i(a, g, a, b)
	inline val agaa get() = Vector4i(a, g, a, a)
	inline val abrr get() = Vector4i(a, b, r, r)
	inline val abrg get() = Vector4i(a, b, r, g)
	inline val abrb get() = Vector4i(a, b, r, b)
	inline val abra get() = Vector4i(a, b, r, a)
	inline val abgr get() = Vector4i(a, b, g, r)
	inline val abgg get() = Vector4i(a, b, g, g)
	inline val abgb get() = Vector4i(a, b, g, b)
	inline val abga get() = Vector4i(a, b, g, a)
	inline val abbr get() = Vector4i(a, b, b, r)
	inline val abbg get() = Vector4i(a, b, b, g)
	inline val abbb get() = Vector4i(a, b, b, b)
	inline val abba get() = Vector4i(a, b, b, a)
	inline val abar get() = Vector4i(a, b, a, r)
	inline val abag get() = Vector4i(a, b, a, g)
	inline val abab get() = Vector4i(a, b, a, b)
	inline val abaa get() = Vector4i(a, b, a, a)
	inline val aarr get() = Vector4i(a, a, r, r)
	inline val aarg get() = Vector4i(a, a, r, g)
	inline val aarb get() = Vector4i(a, a, r, b)
	inline val aara get() = Vector4i(a, a, r, a)
	inline val aagr get() = Vector4i(a, a, g, r)
	inline val aagg get() = Vector4i(a, a, g, g)
	inline val aagb get() = Vector4i(a, a, g, b)
	inline val aaga get() = Vector4i(a, a, g, a)
	inline val aabr get() = Vector4i(a, a, b, r)
	inline val aabg get() = Vector4i(a, a, b, g)
	inline val aabb get() = Vector4i(a, a, b, b)
	inline val aaba get() = Vector4i(a, a, b, a)
	inline val aaar get() = Vector4i(a, a, a, r)
	inline val aaag get() = Vector4i(a, a, a, g)
	inline val aaab get() = Vector4i(a, a, a, b)
	inline val aaaa get() = Vector4i(a, a, a, a)
	inline val ss get() = Vector2i(s, s)
	inline val st get() = Vector2i(s, t)
	inline val sp get() = Vector2i(s, p)
	inline val sq get() = Vector2i(s, q)
	inline val ts get() = Vector2i(t, s)
	inline val tt get() = Vector2i(t, t)
	inline val tp get() = Vector2i(t, p)
	inline val tq get() = Vector2i(t, q)
	inline val ps get() = Vector2i(p, s)
	inline val pt get() = Vector2i(p, t)
	inline val pp get() = Vector2i(p, p)
	inline val pq get() = Vector2i(p, q)
	inline val qs get() = Vector2i(q, s)
	inline val qt get() = Vector2i(q, t)
	inline val qp get() = Vector2i(q, p)
	inline val qq get() = Vector2i(q, q)
	inline val sss get() = Vector3i(s, s, s)
	inline val sst get() = Vector3i(s, s, t)
	inline val ssp get() = Vector3i(s, s, p)
	inline val ssq get() = Vector3i(s, s, q)
	inline val sts get() = Vector3i(s, t, s)
	inline val stt get() = Vector3i(s, t, t)
	inline val stp get() = Vector3i(s, t, p)
	inline val stq get() = Vector3i(s, t, q)
	inline val sps get() = Vector3i(s, p, s)
	inline val spt get() = Vector3i(s, p, t)
	inline val spp get() = Vector3i(s, p, p)
	inline val spq get() = Vector3i(s, p, q)
	inline val sqs get() = Vector3i(s, q, s)
	inline val sqt get() = Vector3i(s, q, t)
	inline val sqp get() = Vector3i(s, q, p)
	inline val sqq get() = Vector3i(s, q, q)
	inline val tss get() = Vector3i(t, s, s)
	inline val tst get() = Vector3i(t, s, t)
	inline val tsp get() = Vector3i(t, s, p)
	inline val tsq get() = Vector3i(t, s, q)
	inline val tts get() = Vector3i(t, t, s)
	inline val ttt get() = Vector3i(t, t, t)
	inline val ttp get() = Vector3i(t, t, p)
	inline val ttq get() = Vector3i(t, t, q)
	inline val tps get() = Vector3i(t, p, s)
	inline val tpt get() = Vector3i(t, p, t)
	inline val tpp get() = Vector3i(t, p, p)
	inline val tpq get() = Vector3i(t, p, q)
	inline val tqs get() = Vector3i(t, q, s)
	inline val tqt get() = Vector3i(t, q, t)
	inline val tqp get() = Vector3i(t, q, p)
	inline val tqq get() = Vector3i(t, q, q)
	inline val pss get() = Vector3i(p, s, s)
	inline val pst get() = Vector3i(p, s, t)
	inline val psp get() = Vector3i(p, s, p)
	inline val psq get() = Vector3i(p, s, q)
	inline val pts get() = Vector3i(p, t, s)
	inline val ptt get() = Vector3i(p, t, t)
	inline val ptp get() = Vector3i(p, t, p)
	inline val ptq get() = Vector3i(p, t, q)
	inline val pps get() = Vector3i(p, p, s)
	inline val ppt get() = Vector3i(p, p, t)
	inline val ppp get() = Vector3i(p, p, p)
	inline val ppq get() = Vector3i(p, p, q)
	inline val pqs get() = Vector3i(p, q, s)
	inline val pqt get() = Vector3i(p, q, t)
	inline val pqp get() = Vector3i(p, q, p)
	inline val pqq get() = Vector3i(p, q, q)
	inline val qss get() = Vector3i(q, s, s)
	inline val qst get() = Vector3i(q, s, t)
	inline val qsp get() = Vector3i(q, s, p)
	inline val qsq get() = Vector3i(q, s, q)
	inline val qts get() = Vector3i(q, t, s)
	inline val qtt get() = Vector3i(q, t, t)
	inline val qtp get() = Vector3i(q, t, p)
	inline val qtq get() = Vector3i(q, t, q)
	inline val qps get() = Vector3i(q, p, s)
	inline val qpt get() = Vector3i(q, p, t)
	inline val qpp get() = Vector3i(q, p, p)
	inline val qpq get() = Vector3i(q, p, q)
	inline val qqs get() = Vector3i(q, q, s)
	inline val qqt get() = Vector3i(q, q, t)
	inline val qqp get() = Vector3i(q, q, p)
	inline val qqq get() = Vector3i(q, q, q)
	inline val ssss get() = Vector4i(s, s, s, s)
	inline val ssst get() = Vector4i(s, s, s, t)
	inline val sssp get() = Vector4i(s, s, s, p)
	inline val sssq get() = Vector4i(s, s, s, q)
	inline val ssts get() = Vector4i(s, s, t, s)
	inline val sstt get() = Vector4i(s, s, t, t)
	inline val sstp get() = Vector4i(s, s, t, p)
	inline val sstq get() = Vector4i(s, s, t, q)
	inline val ssps get() = Vector4i(s, s, p, s)
	inline val sspt get() = Vector4i(s, s, p, t)
	inline val sspp get() = Vector4i(s, s, p, p)
	inline val sspq get() = Vector4i(s, s, p, q)
	inline val ssqs get() = Vector4i(s, s, q, s)
	inline val ssqt get() = Vector4i(s, s, q, t)
	inline val ssqp get() = Vector4i(s, s, q, p)
	inline val ssqq get() = Vector4i(s, s, q, q)
	inline val stss get() = Vector4i(s, t, s, s)
	inline val stst get() = Vector4i(s, t, s, t)
	inline val stsp get() = Vector4i(s, t, s, p)
	inline val stsq get() = Vector4i(s, t, s, q)
	inline val stts get() = Vector4i(s, t, t, s)
	inline val sttt get() = Vector4i(s, t, t, t)
	inline val sttp get() = Vector4i(s, t, t, p)
	inline val sttq get() = Vector4i(s, t, t, q)
	inline val stps get() = Vector4i(s, t, p, s)
	inline val stpt get() = Vector4i(s, t, p, t)
	inline val stpp get() = Vector4i(s, t, p, p)
	inline val stpq get() = Vector4i(s, t, p, q)
	inline val stqs get() = Vector4i(s, t, q, s)
	inline val stqt get() = Vector4i(s, t, q, t)
	inline val stqp get() = Vector4i(s, t, q, p)
	inline val stqq get() = Vector4i(s, t, q, q)
	inline val spss get() = Vector4i(s, p, s, s)
	inline val spst get() = Vector4i(s, p, s, t)
	inline val spsp get() = Vector4i(s, p, s, p)
	inline val spsq get() = Vector4i(s, p, s, q)
	inline val spts get() = Vector4i(s, p, t, s)
	inline val sptt get() = Vector4i(s, p, t, t)
	inline val sptp get() = Vector4i(s, p, t, p)
	inline val sptq get() = Vector4i(s, p, t, q)
	inline val spps get() = Vector4i(s, p, p, s)
	inline val sppt get() = Vector4i(s, p, p, t)
	inline val sppp get() = Vector4i(s, p, p, p)
	inline val sppq get() = Vector4i(s, p, p, q)
	inline val spqs get() = Vector4i(s, p, q, s)
	inline val spqt get() = Vector4i(s, p, q, t)
	inline val spqp get() = Vector4i(s, p, q, p)
	inline val spqq get() = Vector4i(s, p, q, q)
	inline val sqss get() = Vector4i(s, q, s, s)
	inline val sqst get() = Vector4i(s, q, s, t)
	inline val sqsp get() = Vector4i(s, q, s, p)
	inline val sqsq get() = Vector4i(s, q, s, q)
	inline val sqts get() = Vector4i(s, q, t, s)
	inline val sqtt get() = Vector4i(s, q, t, t)
	inline val sqtp get() = Vector4i(s, q, t, p)
	inline val sqtq get() = Vector4i(s, q, t, q)
	inline val sqps get() = Vector4i(s, q, p, s)
	inline val sqpt get() = Vector4i(s, q, p, t)
	inline val sqpp get() = Vector4i(s, q, p, p)
	inline val sqpq get() = Vector4i(s, q, p, q)
	inline val sqqs get() = Vector4i(s, q, q, s)
	inline val sqqt get() = Vector4i(s, q, q, t)
	inline val sqqp get() = Vector4i(s, q, q, p)
	inline val sqqq get() = Vector4i(s, q, q, q)
	inline val tsss get() = Vector4i(t, s, s, s)
	inline val tsst get() = Vector4i(t, s, s, t)
	inline val tssp get() = Vector4i(t, s, s, p)
	inline val tssq get() = Vector4i(t, s, s, q)
	inline val tsts get() = Vector4i(t, s, t, s)
	inline val tstt get() = Vector4i(t, s, t, t)
	inline val tstp get() = Vector4i(t, s, t, p)
	inline val tstq get() = Vector4i(t, s, t, q)
	inline val tsps get() = Vector4i(t, s, p, s)
	inline val tspt get() = Vector4i(t, s, p, t)
	inline val tspp get() = Vector4i(t, s, p, p)
	inline val tspq get() = Vector4i(t, s, p, q)
	inline val tsqs get() = Vector4i(t, s, q, s)
	inline val tsqt get() = Vector4i(t, s, q, t)
	inline val tsqp get() = Vector4i(t, s, q, p)
	inline val tsqq get() = Vector4i(t, s, q, q)
	inline val ttss get() = Vector4i(t, t, s, s)
	inline val ttst get() = Vector4i(t, t, s, t)
	inline val ttsp get() = Vector4i(t, t, s, p)
	inline val ttsq get() = Vector4i(t, t, s, q)
	inline val ttts get() = Vector4i(t, t, t, s)
	inline val tttt get() = Vector4i(t, t, t, t)
	inline val tttp get() = Vector4i(t, t, t, p)
	inline val tttq get() = Vector4i(t, t, t, q)
	inline val ttps get() = Vector4i(t, t, p, s)
	inline val ttpt get() = Vector4i(t, t, p, t)
	inline val ttpp get() = Vector4i(t, t, p, p)
	inline val ttpq get() = Vector4i(t, t, p, q)
	inline val ttqs get() = Vector4i(t, t, q, s)
	inline val ttqt get() = Vector4i(t, t, q, t)
	inline val ttqp get() = Vector4i(t, t, q, p)
	inline val ttqq get() = Vector4i(t, t, q, q)
	inline val tpss get() = Vector4i(t, p, s, s)
	inline val tpst get() = Vector4i(t, p, s, t)
	inline val tpsp get() = Vector4i(t, p, s, p)
	inline val tpsq get() = Vector4i(t, p, s, q)
	inline val tpts get() = Vector4i(t, p, t, s)
	inline val tptt get() = Vector4i(t, p, t, t)
	inline val tptp get() = Vector4i(t, p, t, p)
	inline val tptq get() = Vector4i(t, p, t, q)
	inline val tpps get() = Vector4i(t, p, p, s)
	inline val tppt get() = Vector4i(t, p, p, t)
	inline val tppp get() = Vector4i(t, p, p, p)
	inline val tppq get() = Vector4i(t, p, p, q)
	inline val tpqs get() = Vector4i(t, p, q, s)
	inline val tpqt get() = Vector4i(t, p, q, t)
	inline val tpqp get() = Vector4i(t, p, q, p)
	inline val tpqq get() = Vector4i(t, p, q, q)
	inline val tqss get() = Vector4i(t, q, s, s)
	inline val tqst get() = Vector4i(t, q, s, t)
	inline val tqsp get() = Vector4i(t, q, s, p)
	inline val tqsq get() = Vector4i(t, q, s, q)
	inline val tqts get() = Vector4i(t, q, t, s)
	inline val tqtt get() = Vector4i(t, q, t, t)
	inline val tqtp get() = Vector4i(t, q, t, p)
	inline val tqtq get() = Vector4i(t, q, t, q)
	inline val tqps get() = Vector4i(t, q, p, s)
	inline val tqpt get() = Vector4i(t, q, p, t)
	inline val tqpp get() = Vector4i(t, q, p, p)
	inline val tqpq get() = Vector4i(t, q, p, q)
	inline val tqqs get() = Vector4i(t, q, q, s)
	inline val tqqt get() = Vector4i(t, q, q, t)
	inline val tqqp get() = Vector4i(t, q, q, p)
	inline val tqqq get() = Vector4i(t, q, q, q)
	inline val psss get() = Vector4i(p, s, s, s)
	inline val psst get() = Vector4i(p, s, s, t)
	inline val pssp get() = Vector4i(p, s, s, p)
	inline val pssq get() = Vector4i(p, s, s, q)
	inline val psts get() = Vector4i(p, s, t, s)
	inline val pstt get() = Vector4i(p, s, t, t)
	inline val pstp get() = Vector4i(p, s, t, p)
	inline val pstq get() = Vector4i(p, s, t, q)
	inline val psps get() = Vector4i(p, s, p, s)
	inline val pspt get() = Vector4i(p, s, p, t)
	inline val pspp get() = Vector4i(p, s, p, p)
	inline val pspq get() = Vector4i(p, s, p, q)
	inline val psqs get() = Vector4i(p, s, q, s)
	inline val psqt get() = Vector4i(p, s, q, t)
	inline val psqp get() = Vector4i(p, s, q, p)
	inline val psqq get() = Vector4i(p, s, q, q)
	inline val ptss get() = Vector4i(p, t, s, s)
	inline val ptst get() = Vector4i(p, t, s, t)
	inline val ptsp get() = Vector4i(p, t, s, p)
	inline val ptsq get() = Vector4i(p, t, s, q)
	inline val ptts get() = Vector4i(p, t, t, s)
	inline val pttt get() = Vector4i(p, t, t, t)
	inline val pttp get() = Vector4i(p, t, t, p)
	inline val pttq get() = Vector4i(p, t, t, q)
	inline val ptps get() = Vector4i(p, t, p, s)
	inline val ptpt get() = Vector4i(p, t, p, t)
	inline val ptpp get() = Vector4i(p, t, p, p)
	inline val ptpq get() = Vector4i(p, t, p, q)
	inline val ptqs get() = Vector4i(p, t, q, s)
	inline val ptqt get() = Vector4i(p, t, q, t)
	inline val ptqp get() = Vector4i(p, t, q, p)
	inline val ptqq get() = Vector4i(p, t, q, q)
	inline val ppss get() = Vector4i(p, p, s, s)
	inline val ppst get() = Vector4i(p, p, s, t)
	inline val ppsp get() = Vector4i(p, p, s, p)
	inline val ppsq get() = Vector4i(p, p, s, q)
	inline val ppts get() = Vector4i(p, p, t, s)
	inline val pptt get() = Vector4i(p, p, t, t)
	inline val pptp get() = Vector4i(p, p, t, p)
	inline val pptq get() = Vector4i(p, p, t, q)
	inline val ppps get() = Vector4i(p, p, p, s)
	inline val pppt get() = Vector4i(p, p, p, t)
	inline val pppp get() = Vector4i(p, p, p, p)
	inline val pppq get() = Vector4i(p, p, p, q)
	inline val ppqs get() = Vector4i(p, p, q, s)
	inline val ppqt get() = Vector4i(p, p, q, t)
	inline val ppqp get() = Vector4i(p, p, q, p)
	inline val ppqq get() = Vector4i(p, p, q, q)
	inline val pqss get() = Vector4i(p, q, s, s)
	inline val pqst get() = Vector4i(p, q, s, t)
	inline val pqsp get() = Vector4i(p, q, s, p)
	inline val pqsq get() = Vector4i(p, q, s, q)
	inline val pqts get() = Vector4i(p, q, t, s)
	inline val pqtt get() = Vector4i(p, q, t, t)
	inline val pqtp get() = Vector4i(p, q, t, p)
	inline val pqtq get() = Vector4i(p, q, t, q)
	inline val pqps get() = Vector4i(p, q, p, s)
	inline val pqpt get() = Vector4i(p, q, p, t)
	inline val pqpp get() = Vector4i(p, q, p, p)
	inline val pqpq get() = Vector4i(p, q, p, q)
	inline val pqqs get() = Vector4i(p, q, q, s)
	inline val pqqt get() = Vector4i(p, q, q, t)
	inline val pqqp get() = Vector4i(p, q, q, p)
	inline val pqqq get() = Vector4i(p, q, q, q)
	inline val qsss get() = Vector4i(q, s, s, s)
	inline val qsst get() = Vector4i(q, s, s, t)
	inline val qssp get() = Vector4i(q, s, s, p)
	inline val qssq get() = Vector4i(q, s, s, q)
	inline val qsts get() = Vector4i(q, s, t, s)
	inline val qstt get() = Vector4i(q, s, t, t)
	inline val qstp get() = Vector4i(q, s, t, p)
	inline val qstq get() = Vector4i(q, s, t, q)
	inline val qsps get() = Vector4i(q, s, p, s)
	inline val qspt get() = Vector4i(q, s, p, t)
	inline val qspp get() = Vector4i(q, s, p, p)
	inline val qspq get() = Vector4i(q, s, p, q)
	inline val qsqs get() = Vector4i(q, s, q, s)
	inline val qsqt get() = Vector4i(q, s, q, t)
	inline val qsqp get() = Vector4i(q, s, q, p)
	inline val qsqq get() = Vector4i(q, s, q, q)
	inline val qtss get() = Vector4i(q, t, s, s)
	inline val qtst get() = Vector4i(q, t, s, t)
	inline val qtsp get() = Vector4i(q, t, s, p)
	inline val qtsq get() = Vector4i(q, t, s, q)
	inline val qtts get() = Vector4i(q, t, t, s)
	inline val qttt get() = Vector4i(q, t, t, t)
	inline val qttp get() = Vector4i(q, t, t, p)
	inline val qttq get() = Vector4i(q, t, t, q)
	inline val qtps get() = Vector4i(q, t, p, s)
	inline val qtpt get() = Vector4i(q, t, p, t)
	inline val qtpp get() = Vector4i(q, t, p, p)
	inline val qtpq get() = Vector4i(q, t, p, q)
	inline val qtqs get() = Vector4i(q, t, q, s)
	inline val qtqt get() = Vector4i(q, t, q, t)
	inline val qtqp get() = Vector4i(q, t, q, p)
	inline val qtqq get() = Vector4i(q, t, q, q)
	inline val qpss get() = Vector4i(q, p, s, s)
	inline val qpst get() = Vector4i(q, p, s, t)
	inline val qpsp get() = Vector4i(q, p, s, p)
	inline val qpsq get() = Vector4i(q, p, s, q)
	inline val qpts get() = Vector4i(q, p, t, s)
	inline val qptt get() = Vector4i(q, p, t, t)
	inline val qptp get() = Vector4i(q, p, t, p)
	inline val qptq get() = Vector4i(q, p, t, q)
	inline val qpps get() = Vector4i(q, p, p, s)
	inline val qppt get() = Vector4i(q, p, p, t)
	inline val qppp get() = Vector4i(q, p, p, p)
	inline val qppq get() = Vector4i(q, p, p, q)
	inline val qpqs get() = Vector4i(q, p, q, s)
	inline val qpqt get() = Vector4i(q, p, q, t)
	inline val qpqp get() = Vector4i(q, p, q, p)
	inline val qpqq get() = Vector4i(q, p, q, q)
	inline val qqss get() = Vector4i(q, q, s, s)
	inline val qqst get() = Vector4i(q, q, s, t)
	inline val qqsp get() = Vector4i(q, q, s, p)
	inline val qqsq get() = Vector4i(q, q, s, q)
	inline val qqts get() = Vector4i(q, q, t, s)
	inline val qqtt get() = Vector4i(q, q, t, t)
	inline val qqtp get() = Vector4i(q, q, t, p)
	inline val qqtq get() = Vector4i(q, q, t, q)
	inline val qqps get() = Vector4i(q, q, p, s)
	inline val qqpt get() = Vector4i(q, q, p, t)
	inline val qqpp get() = Vector4i(q, q, p, p)
	inline val qqpq get() = Vector4i(q, q, p, q)
	inline val qqqs get() = Vector4i(q, q, q, s)
	inline val qqqt get() = Vector4i(q, q, q, t)
	inline val qqqp get() = Vector4i(q, q, q, p)
	inline val qqqq get() = Vector4i(q, q, q, q)

	companion object {
		@JvmField val ZERO = Vector4i()
		@JvmField val POSITIVE_X = Vector4i(x = 1)
		@JvmField val NEGATIVE_X = Vector4i(x = -1)
		@JvmField val POSITIVE_Y = Vector4i(y = 1)
		@JvmField val NEGATIVE_Y = Vector4i(y = -1)
		@JvmField val POSITIVE_Z = Vector4i(z = 1)
		@JvmField val NEGATIVE_Z = Vector4i(z = -1)
		@JvmField val POSITIVE_W = Vector4i(w = 1)
		@JvmField val NEGATIVE_W = Vector4i(w = -1)
	}
}

operator fun Int.times(other: Vector4i): Vector4i {
	return Vector4i(this * other.x, this * other.y, this * other.z, this * other.w)
}

operator fun Int.div(other: Vector4i): Vector4i {
	return Vector4i(this / other.x, this / other.y, this / other.z, this / other.w)
}

/**
 * Mutable 3D Vector, representing three-dimensional coordinates as [Int]s
 *
 * It can be safely passed around to methods which expect immutable [Vector4i], **AND** do not store it,
 * as mutable operations are separated from immutable ones
 */
open class MutableVector4i(
	override var x: Int = 0,
	override var y: Int = 0,
	override var z: Int = 0,
	override var w: Int = 0,
) : Vector4i(x, y, z, w), IMutableIntVector<MutableVector4i>, IMutableWholeVector<MutableVector4i>, IMutableVector<MutableVector4i, Vector4i>, IMatrixSetterInt {
	constructor(input: IStruct2i, z: Int = 0, w: Int = 0) : this(input.component1(), input.component2(), z, w)
	constructor(x: Int, input: IStruct2i, w: Int = 0) : this(x, input.component1(), input.component2(), w)
	constructor(x: Int, y: Int, input: IStruct2i) : this(x, y, input.component1(), input.component2())

	constructor(input: IStruct3i, w: Int = 0) : this(input.component1(), input.component2(), input.component3(), w)
	constructor(x: Int, input: IStruct3i) : this(x, input.component1(), input.component2(), input.component3())
	constructor(input: IStruct4i) : this(input.component1(), input.component2(), input.component3(), input.component4())

	override fun set(column: Int, row: Int, value: Int) {
		if (column != 0) {
			throw IndexOutOfBoundsException("Vectors are 1 column matrices ($column given)")
		}

		when (row) {
			0 -> x = value
			1 -> y = value
			2 -> z = value
			3 -> w = value
			else -> throw IndexOutOfBoundsException("Row out of bounds: $row")
		}
	}

	override fun equals(other: Any?): Boolean {
		if (this === other) return true
		if (javaClass != other?.javaClass) return false

		other as MutableVector4i

		if (x != other.x) return false
		if (y != other.y) return false
		if (z != other.z) return false
		if (w != other.w) return false

		return true
	}

	override var r by this::x
	override var g by this::y
	override var b by this::z
	override var a by this::w

	override var s by this::x
	override var t by this::y
	override var p by this::z
	override var q by this::w

	override fun plusMut(other: Vector4i): MutableVector4i {
		x += other.x
		y += other.y
		z += other.z
		z += other.z
		return this
	}

	override fun minusMut(other: Vector4i): MutableVector4i {
		x -= other.x
		y -= other.y
		z -= other.z
		z -= other.z
		return this
	}

	override fun timesMut(other: Vector4i): MutableVector4i {
		x *= other.x
		y *= other.y
		z *= other.z
		z *= other.z
		return this
	}

	override fun divMut(other: Vector4i): MutableVector4i {
		x /= other.x
		y /= other.y
		z /= other.z
		z /= other.z
		return this
	}

	override fun timesMut(other: Int): MutableVector4i {
		x *= other
		y *= other
		z *= other
		z *= other
		return this
	}

	override fun divMut(other: Int): MutableVector4i {
		x /= other
		y /= other
		z /= other
		z /= other
		return this
	}
}
