package ru.dbotthepony.kstarbound.world.entities

import ru.dbotthepony.kbox2d.api.BodyDef
import ru.dbotthepony.kbox2d.api.BodyType
import ru.dbotthepony.kvector.util2d.AABB
import ru.dbotthepony.kvector.vector.ndouble.Vector2d

enum class CollisionResolution {
	STOP,
	BOUNCE,
	PUSH,
	SLIDE,
}

abstract class MovementController<T : IEntity>(val entity: T) {
	val world = entity.world
	open var position by entity::position
	open var angle by entity::angle

	protected val body by lazy {
		world.physics.createBody(BodyDef(
			position = position,
			angle = angle,
			type = BodyType.DYNAMIC
		))
	}

	open val velocity get() = body.linearVelocity

	/**
	 * Returns whenever are we contacting something below us
	 */
	open val onGround: Boolean get() {
		for (contact in body.contactEdgeIterator) {
			if (contact.contact.manifold.localNormal.dot(world.gravity.normalized) >= 0.97) {
				return true
			}
		}

		return false
	}

	/**
	 * World space AABB, by default returning combined AABB of physical body.
	 */
	open val worldAABB: AABB get() {
		return body.worldSpaceAABB
	}

	open fun thinkPhysics(delta: Double) {
		mutePositionChanged = true
		position = body.position
		angle = body.angle
		mutePositionChanged = false
	}

	protected open fun onTouchSurface(velocity: Vector2d, normal: Vector2d) {
		entity.onTouchSurface(velocity, normal)
	}

	protected var mutePositionChanged = false

	open fun notifyPositionChanged() {
		if (mutePositionChanged) {
			return
		}

		body.setTransform(entity.position, entity.angle)
	}
}

/**
 * Movement controller for logical entities, which does nothing.
 */
class LogicalMovementController(entity: Entity) : MovementController<Entity>(entity) {
	override val worldAABB: AABB get() = AABB(position, position)

	// Dummies never touch anything, since they don't have Box2D body
	override val onGround: Boolean = false
	override val velocity: Vector2d = Vector2d.ZERO

	override fun thinkPhysics(delta: Double) {
		// no-op
	}

	override fun notifyPositionChanged() {
		// no-op
	}

	companion object {
		private val DUMMY_AABB = AABB.rectangle(Vector2d.ZERO, 0.1, 0.1)
	}
}
